# MUD Debug Tools.
# Copyright 2021 runphase.com .  All rights reserved
#
from pdb import Pdb, set_trace as enter_debugger, runcall as debugCall

# Basic debugging implementation.
def breakOn(function):
    def debugWrapper(*args, **kwd):
        return debugCall(function, *args, **kwd)

    debugWrapper.__name__ = getattr(function, '__name__', '')
    debugWrapper.__doc__ = getattr(function, '__doc__', '')
    return debugWrapper

def traceOn(function):
    def traceWrapper(*args, **kwd):
        enter_debugger()
        return function(*args, **kwd)

    traceWrapper.__name__ = getattr(function, '__name__', '')
    traceWrapper.__doc__ = getattr(function, '__doc__', '')
    return traceWrapper

def remoteBreak(function):
    def remoteDebugWrapper(*args, **kwd):
        from rpdb import Rdb
        return Rdb().runcall(function, *args, **kwd)

    remoteDebugWrapper.__name__ = getattr(function, '__name__', '')
    remoteDebugWrapper.__doc__ = getattr(function, '__doc__', '')
    return remoteDebugWrapper

def remoteTrace(function):
    def remoteTraceWrapper(*args, **kwd):
        from rpdb import Rdb
        Rdb().set_trace()
        return function(*args, **kwd)

    remoteTraceWrapper.__name__ = getattr(function, '__name__', '')
    remoteTraceWrapper.__doc__ = getattr(function, '__doc__', '')
    return remoteTraceWrapper

# Hard-level Trace Tools.
def trace(*args, **kwd):
    from stuphos.system.api import mudlog
    scopeName = kwd.get('scopeName')

    args = ' '.join(map(str, args))
    if scopeName:
        mudlog('%s (%s)' % (args, scopeName))
    else:
        mudlog(args)

def traceInline(*args, **kwd):
    from ph import enqueueHeartbeatTask
    enqueueHeartbeatTask(trace, *args)

HARDLOG_FILENAME_FORMAT = '../log/hardlog.%s.txt'
def hardlog(*args, **kwd):
    topicName = kwd.get('topicName', 'main')

    try: open(HARDLOG_FILENAME_FORMAT % topicName, 'a').write(' '.join(map(str, args)))
    except IOError as e:
        from errno import ENOENT
        if e.args[0] != ENOENT:
            raise e

asBuiltin(trace)
asBuiltin(traceInline)
asBuiltin(hardlog)

# Disabled:
##    # Programmed for Pdb Overseer.
##    from pdb import Pdb, set_trace as enter_debugger, runcall as debugCall
##    import sys
##
##    class DebugCompartment(Pdb):
##        # Handles
##        def __init__(self, breakpoints = None):
##            Pdb.__init__(self)
##            self.load_breakpoints(breakpoints)
##
##        def load_breakpoints(self, breakpoints):
##            if breakpoints:
##                for (filename, lineno) in parse_breakpoints(breakpoints):
##                    self.perform_set_break(filename, lineno)
##
##        def perform_set_break(self, filename, lineno):
##            # High-level breakset.
##            outcome = self.set_break(filename, lineno)
##            if outcome:
##                print outcome
##
##            bp = self.get_breaks(filename, lineno)
##            if bp:
##                bp = bp[-1]
##                print 'Breakpoint %d at %s:%d' % (bp.number, bp.file, bp.line)
##
##        def __repr__(self):
##            from pprint import pformat as pf
##            cname = self.__class__.__name__
##            return '<%s %r>' % (cname, ('\n%s' % (len(cname) + 2)).join(pf(self.breaks).split('\n')))
##
##    def stream_breakpoints(breakpoints):
##        if callable(getattr(breakpoints, 'readlines', None)):
##            def stream():
##                for bp in breakpoints.readlines():
##                    # Trim comments.
##                    pos = bp.find('#')
##                    if pos >= 0:
##                        bp = bp[:pos]
##
##                    # Cheap strchr?
##                    if any(c not in (' \t\b\n\r') for c in bp):
##                        yield bp.rstrip()
##
##            return stream()
##        return breakpoints
##
##    def parse_breakpoints(breakpoints):
##        # Generator.
##        for bp in stream_breakpoints(breakpoints):
##            (filename, lineno) = evaluate_target(bp)
##            if filename:
##                yield (filename, lineno)
##
##    from types import FunctionType as function
##    def evaluate_target(bp):
##        # Parsing and resolution phase.
##        if type(bp) is function:
##            return (bp.func_code.co_filename, bp.fun_code.co_firstlineno)
##        if type(bp) in (tuple, list) and len(bp) is 2 and \
##           type(bp[0]) is str and type(bp[1]) is int:
##            return bp
##
##        (filename, lineno) = ('', '')
##
##        colon = bp.find(':')
##        if colon >= 0:
##            filename = bp[:colon]
##            lineno = int(bp[colon+1:])
##        else:
##            # Search for qualified module/class/method filename/lineno.
##            parts = bp.split('.')
##            try:
##                if len(parts) == 1:
##                    target = eval(parts[0])
##                else:
##                    b = parts[0]
##
##                    target = __import__(b)
##                    z = False # Considered all imports requisited.
##                    for i in xrange(1, len(parts)):
##                        n = parts[i]
##                        if not z:
##                            b = '%s.%s' % (b, n)
##                            try: __import__(b)
##                            except ImportError:
##                                z = True
##
##                        target = getattr(target, n)
##
##                # Dereference target function.
##                code = target.func_code
##
##                # filename = self.canonic(code.co_filename)
##                filename = code.co_filename
##                lineno = code.co_firstlineno
##
##                return (filename, lineno)
##
##            except (ImportError, AttributeError), e:
##                print str(e)
##
##        return (filename, lineno)
##
##    def openNone(*args, **kwd):
##        try: return open(*args, **kwd)
##        except IOError, e:
##            from errno import ENOENT
##            if e.errno != ENOENT:
##                raise
##
##    class MainFrame(DebugCompartment):
##        BREAKPOINTS_FILE = 'python/.breakpoints'
##        MUDFRAME = 'runMUD'
##
##        def __init__(self):
##            DebugCompartment.__init__(self, openNone(self.BREAKPOINTS_FILE))
##            self.set_mud()
##
##        def set_mud(self, name = None):
##            # Start tracing the main frame, immediately, but only stopping for breakpoints.
##            name = name or self.MUDFRAME
##
##            import mud.tools.frames
##            frame = mud.tools.frames.find(name = name, ascending = True)
##
##            self.reset()
##            self.set_trace(frame)
##
##            self.set_continue()
##            # self.set_next()
##            # self.set_return()
##
##            sys.settrace(self.trace_dispatch)
##
##    # Another implementation.
##    DEFAULT_BREAKPOINT_STATE = True
##    _breakpoint_registry = {None: False}
##
##    @asBuiltin
##    def breakOn(point, onStart = DEFAULT_BREAKPOINT_STATE):
##        if type(point) is str:
##            def makeBreakpoint(function):
##                if point == 'this' or point == 'regardless':
##                    breakpoint = getFunctionIdentityBreakpoint(function)
##                else:
##                    breakpoint = point
##
##                function.breakpoint = breakpoint
##                if onStart or point == 'regardless':
##                    enableBreakpoint(breakpoint)
##
##                def debugCall(*args, **kwd):
##                    if not getBreakStatus(function):
##                        return function(*args, **kwd)
##
##                    return runcall(function, *args, **kwd)
##
##                debugCall.__name__ = function.__name__
##                debugCall.__doc__ = function.__doc__
##                debugCall.breakpoint = breakpoint
##
##                return debugCall
##            return makeBreakpoint
##
##        if not callable(point):
##            return point
##
##        def debugCall(*args, **kwd):
##            return runcall(point, *args, **kwd)
##
##        try: debugCall.__name__ = point.__name__
##        except AttributeError: pass
##        try: debugCall.__doc__ = point.__doc__
##        except AttributeError: pass
##
##        return debugCall
##
##    @asBuiltin
##    def enableBreakpoint(object):
##        _breakpoint_registry[getObjectBreakpoint(object)] = True
##
##    @asBuiltin
##    def disableBreakpoint(object):
##        _breakpoint_registry[getObjectBreakpoint(object)] = False
##
##    def getFunctionIdentityBreakpoint(function):
##        return (function.func_code.co_filename, function.func_code.co_firstlineno)
##
##    def getObjectBreakpoint(object):
##        # The breakpoint itself.
##        if type(object) in (str, tuple, list):
##            return object
##
##        try: return object.breakpoint
##        except AttributeError:
##            if type(object) is FunctionType:
##                return getFunctionIdentityBreakpoint(object)
##
##    def isBreakpointEnabled(name):
##        return bool(_breakpoint_registry.get(name, False))
##    def getBreakStatus(object):
##        return isBreakpointEnabled(getObjectBreakpoint(object))
##
##    ##    __builtins__.update(dict(breakOn = breakOn,
##    ##                             debug = enter_debugger,
##    ##                             enter_debugger = enter_debugger))
##
##    def dbgpClientBrk():
##        # Komodo/xdebug
##        from dbgp.client import brk
##        brk(port = 9000)
##
##    asBuiltin(dbgpClientBrk)
##
##    def test():
##        return test
##
##    if __name__ == '__main__':
##        from pdb import runcall
##        print runcall(mainFrame)



# Miscellaneous Runtime Tools.
#
# Note: These things aren't really used anymore.
# (This stuff can come from common tools anyway)

# Instrument the Runtime Frame Stack.
from sys import _getframe as thisFrame

def show(order = None):
    for frame in order or stack():
        print(string(frame))

def traceback(frame = None, n = -1):
    frame = frame or thisFrame()
    while frame:
        if n > 0:
            n -= 1
            if n > 0:
                frame = frame.f_back
                continue

        yield frame
        frame = frame.f_back

def stack():
    # Generate return in reverse order from grand-parent calling frame.
    return reversed(list(traceback(n = 2)))

def bottom():
    return list(traceback())[-1]

def string(frame):
    from linecache import getline
    code = frame.f_code
    return '%s:%s:%s%s' % (code.co_name, code.co_filename, code.co_firstlineno,
                           (frame.f_trace and ' (%s)' % frame.f_trace.__name__) or '')

def search(name = None, ascending = False):
    for frame in (ascending and stack() or traceback()):
        if frame.f_code.co_name == name:
            yield frame

def find(*args, **kwd):
    frames = search(*args, **kwd)

    try: return next(frames)
    except StopIteration:
        raise NameError(name)

# Debugger legacy:
##    def set_byframename(name):
##        self.reset()
##        stack = FrameStack.search(name = name, ascending = True)
##
##        try: main = iter(stack).next()
##        except StopIteration:
##            raise NameError(name)
##
##        else:
##            self.set_trace(main)
##            self.set_continue()


# Unused Patterns:
class Bundle:
    # Curry/Delegate Pattern.
    def __init__(self, function, *args, **kwd):
        self.function = function
        self.args = args
        self.kwd = kwd

    def __call__(self, *args, **kwd):
        args = self.args + args
        kwd.update(self.kwd)
        return self.function(*args, **kwd)



# Initialize Management Services for Bootstrap.
from stuphos.etc.tools import isYesValue
from .services import AutoServices

class ManagementServices(AutoServices):
    CONFIG_OBJECT_NAME = 'MUD::Configuration'
    CONFIG_SECTION_NAME = 'Management'

    def loadPentacleServices(self, conf):
        from stuphos.pentacle import partner
        partner.bootPartnerAspect()

    def loadSessionAdapter(self, conf):
        from phsite.network.adapter.sessions import SessionManager
        try: SessionManager.get(create = True) # CreateFacility(SessionManager)
        except RuntimeError as e:
            logOperation(f'[sessions] {e}')

    def loadEmbeddedWebserver(self, conf):
        from stuphos.runtime.architecture.api import LookupObject
        webServiceClass = LookupObject(self.section.get('webserver-object', 'stuphos.webserver.DjangoService'))

        try: webServiceClass.get(create = True) # CreateFacility(webServiceClass)
        except OSError as e:
            logOperation(f'[webserver] {e.__class__.__name__}: {e}')

            from stuphos import logException
            logException(traceback = True)

        # todo: catch error: [Errno 112] Address already in use
        # OSError: [Errno 98] Address already in use


    # Some lesser features.
    def loadSystemShell(self, conf):
        from . import shell
        shell.SystemShell.get(create = True)
        # CreateFacility(shell.SystemShell)

    def loadSubdaemonManager(self, conf):
        from . import subdaemon
        subdaemon.SubDaemonManager.get(create = True)
        # CreateFacility(subdaemon.SubDaemonManager)

    def loadSyslogScanner(self, conf):
        from . import syslog

    def loadCrashHandler(self, conf):
        import sys, signal
        from os import kill, getpid

        def bailOut(signr, frame):
            # Running after a segv does not guarentee success,
            # but if we have enough facility, then we want this data.
            print(system.CallingFrame.Snapshot(), file=sys.stderr)

            # Resend signal with default processor -- force crash.
            signal.signal(signal.SIGSEGV, signal.SIG_DFL)
            kill(getpid(), signal.SIGSEGV)

        signal.signal(signal.SIGSEGV, bailOut)

    def postOn(self, cfg):
        if not isYesValue(cfg.get('embedded-webserver')):
            from stuphos.webserver import djangoConfig
            djangoConfig()


initForCore = ManagementServices(setup_deferred = True)

# def traceToSyslog(frame, pos, instr, args):
#     msg = '%04d %s(%s)' % (pos, getattr(instr, '__name__', '?'),
#                            ', '.join(map(str, args)))

#     import game
#     game.syslog('%s:\n%s' % (msg, '\n'.join(map(repr, frame.task.stack))))

#     # import pdb; pdb.set_trace()
#     # return True # enter (girl) debugger

def runSessionCore(peer, path, activationName, **environ):
    # Structural combination of girl and web.adapter.sessions
    # from stuphos.system.api import game
    from stuphos.kernel import constrainStructureMemory

    try: session = peer.session_ref()
    except AttributeError:
        pass
    else:
        core = runtime[runtime.Agent.System]
        vm = runtime[runtime.System.Engine]

        if session is None or core is None or vm is None:
            def run(function):
                raise RuntimeError('Invalid runtime')
        else:
            try: playerName = peer.avatar.name.lower()
            except AttributeError: progr = None
            else:
                from stuphos.runtime.architecture.api import getWebProgrammer
                progr = getWebProgrammer()(playerName)

            def run(function):
                try:
                    struct = core[path].structure
                    trigger = getattr(struct, activationName) # trigger

                except (AttributeError, KeyError): pass
                except ImportError as e: # no WMF
                    pass # todo: log
                else:
                    try: name = environ.pop('task_name')
                    except KeyError: name = None

                    def initContainer(task):
                        # Whoops: should be using 'environment', set by Trigger.
                        with vm.threadContext(task = task):
                            task.frame.locals['container'] = constrainStructureMemory(task, struct)

                    # debugOn()
                    debug = configuration.Interpreter.debug_session
                    if isYesValue(debug):
                        debug = True

                    # Pass a trigger by naming it via activationName in the struct.
                    if callable(getattr(trigger, '_activate', None)):
                        task = trigger._activate(trigger._module, locals = environ,
                                                 initialize = initContainer,
                                                 programmer = progr,
                                                 audit = debug)

                        task._task.name = name # mud.management.structure.Trigger.Task

                        # game.syslog('operator: %r, audit: %r' % (task._task.operator, task._task.tracing))

                        @task._onComplete
                        def complete(_, exception = None):
                            # p = task._task.frames[-1].procedure
                            # game.syslog('%s\n%s\n%s' % (p.position()-1, p.instructions, task._task.stack))
                            # game.syslog('\n%s' % '\n'.join(map(repr, task._task.stack)))

                            try: result = task._result
                            except IndexError as e:
                                pass # game.syslog('%s' % e)
                            else:
                                try: function(session, result)
                                except:
                                    from stuphos import logException
                                    logException(traceback = True)

        return run



# MUD Player Package.
#   -- stuphmud.server.player

# Interpreter aspect.
# todo: rename to 'mud.interface'
from errno import ENOENT
from pickle import load as load_pickle
from pickle import dump as save_pickle
from types import GeneratorType as generator
import platform

from stuphos.etc import ShowFrame, HandleException, logException
from stuphos.etc import getSystemException, isYesValue
from stuphos.etc import capitalize as getCapitalizedName, stylizedHeader
from stuphos.etc.tools.internet import IPAddressGroup

from stuphos import getConfig, invokeTimeoutForHeartbeat, getBridgeModule
from stuphos.runtime import eventResult, DeclareEvent
from stuphos.runtime.registry import getObject

from .events import TriggerPlayerEvent, Trigger

ENTER_GAME_SCRIPT = None # "'player/interpreter/enterGame'(player)"
DEBUG_ENTER_GAME = False # True # Todo: put this in conf under debug-session/core


#@on.newConnection
def interpret(peer):
    from stuphmud.server.player.shell import ShellI
    peer.interpreter = ShellI(commands = getPlayerCommands(peer),
                              scope = getPlayerScope(peer))

    # Install peer.process_telnet_message
    # Install rich editString and messenger

# Greetings.
def getTitleScreen(peer):
    # Full title.
    title = getConfig('title-screen')
    if title:
        try: return open(title).read()
        except IOError as e:
            if e.args[0] != ENOENT:
                HandleException()

def getGreetings(peer):
    # Return the one-liner.
    greetings = getConfig('greetings')
    if greetings:
        greetings = greetings.strip()
        greetings = greetings.replace('%w', ' ')
        greetings = greetings.replace('%n', '\r\n')
        return greetings

def getTitleAndGreetings(peer):
    t = getTitleScreen(peer)
    g = getGreetings(peer)
    return (t + g) if (t and g) else t if t else g if g else ''

def getGreetingDelay(peer):
    try: return float(getConfig('greeting-delay'))
    except (ValueError, TypeError):
        return None

# mud.api.constants
CON_GET_NAME = 'Get name'

def greetPlayer(peer):
    greetings = getTitleAndGreetings(peer)
    if greetings:
        delay = getGreetingDelay(peer)
        if delay:
            def sendGreeting():
                try:
                    if peer.state == CON_GET_NAME:
                        peer.write(greetings)

                except ValueError:
                    # Peer handle no longer valid.
                    pass

            invokeTimeoutForHeartbeat(delay, sendGreeting)

        else:
            peer.write(greetings)

        return eventResult(True)

def welcomePlayer(peer):
    # New player -- Rename to welcomeNewPlayer?
    pass

class playerActivation(DeclareEvent):
    # An event that can be detached from enterGame.
    Module = getBridgeModule()

def postMessages(peer, type, content):
    try: session = peer.session_ref()
    except AttributeError: pass
    else:
        if session is not None:
            session.postMessages([type, content])

def postJavascript(peer, script):
    postMessages(peer, 'javascript', script)
def portalGo(peer, url):
    postMessages(peer, 'portal', url)


def getDefaultStartRoom(actor):
    return 3001
def getLoadRoom(actor):
    return getattr(actor, 'loadroom', None) # saveroom??

def getFirstValidRoom(rooms):
    # XXX :skip: this is only valid for the circlemud.world impl.
    # import world; table = world.room.table
    # for vnum in rooms:
    #     try: return table[vnum]
    #     except KeyError:
    #         pass

    from stuphos.system.api import world
    lookup = world.room.lookup

    for vnum in rooms:
        if vnum is not None:
            try: return lookup(vnum)
            except KeyError:
                pass

def getPlayerLoadRooms(actor):
    yield getLoadRoom(actor)
    yield getDefaultStartRoom(actor)

def getStartRoom(avatar):
    return getFirstValidRoom(getPlayerLoadRooms(avatar))


def loggedIn(peer, avatar, webAdapter = False, post = None):
    from stuphos.system.api import mudlog, syslog

    mudlog('%s has logged in (web)' % avatar.name) # avatar.level)
    syslog('%s has logged in (web)' % avatar.name)


    # Todo: call player/interpreter/loggedIn, instead of enterGame.
    # That method can initialize location, but the enter-game flow
    # from non-web-adapter main-menu entries needs to be rewritten.
    #
    # Eventually, logged in states differ from enter-game states.

    # Since this isn't done anywhere else during the emerge_internal_peer service call.
    # This should be moved before enterGame.  Do in enterGame.
    avatar.room = getStartRoom(avatar)

    # debugOn()
    # This indirectly calls enterGame (below).
    StuphMUD.EnterGame(peer, avatar, webAdapter = webAdapter, post = post)

    # XXX :skip: Todo: do this by way of enterGame.
    # self.updateCharacter() # selecting 'Character' in player menu will call do-score command.
    # Done via SessionManager.onPlayerActivate
    # if webAdapter:
    #     peer.session.updateLocation()


def enterGame(peer, player, webAdapter = False, post = None):
    from stuphos import getConfig
    from stuphos.etc import isYesValue


    # Prioritize activation handlers before triggered event.
    # debugOn()
    playerActivation(peer, player)

    ##    api = getObject('Web::Extension::API')
    ##    if api is not None:
    ##        player.properties = api.getProperties(player)


    if isYesValue(getConfig('converge-spatial')):
        from spatial.spherical.objects import avatar
        # debugOn()
        player.spatial = avatar._layerZero(player)


    TriggerPlayerEvent(player, 'enter-game')


    from stuphmud.server.adapters import MobileAdapter
    player = MobileAdapter(player) # System perspective


    # Enter game script: task.
    from stuphmud.elemental import PlayerScript

    # Todo: configure these based on webAdapter parameter.
    script = configuration.Interpreter.enter_game_script
    script = script.strip() if isinstance(script, str) else ENTER_GAME_SCRIPT
    debug = configuration.Interpreter.debug_bridge or DEBUG_ENTER_GAME

    if debug not in ['stack', 'console']:
        debug = isYesValue(debug)

    # debugOn()
    scriptComplete = bool(script)
    if script:
        if post:
            def initialize(task):
                # From the web adapter login request.
                task.frame.locals['post'] = task.mapping(*post.items())
        else:
            initialize = None

        # todo: onComplete: scriptComplete = bool(task.stack.pop()[0])
        PlayerScript.evaluateEvent(peer, script,
                                   tracing = debug,
                                   initialize = initialize,
                                   player = player)


    # # Greeting & Enter game script: todo: move into the above evaluateEvent.
    # from stuphos.management import runSessionCore
    # # from stuphos.system.api import syslog

    # name = getConfig('identification', 'Interpreter') or 'identities/{name}'
    # name = name.format(name = peer.avatar.name)
    # # name = 'department/Human Resources/people/' + peer.avatar.name


    # # XXX :skip: Race, because we want panel greeting to display first.
    # complete = runSessionCore(peer, name, 'panelGreeting', player = player,
    #                           task_name = '%s:Greeting' % peer.avatar.name)
    # if complete is not None:
    #     @complete
    #     def run(session, result):
    #         session.postMessages(['html-panel', result])

    # complete = runSessionCore(peer, name, 'enterGame', player = player,
    #                           task_name = '%s:EnterGame(Script)' % peer.avatar.name)
    # if complete is not None:
    #     # syslog('complete-core: %r' % complete)

    #     @complete
    #     def run(session, result):
    #         # print(f'javascript: {result}')
    #         # syslog('CUSTOM SCRIPT:\n%s' % result)
    #         # result is from the vm, meaning it got passed through Stack.append
    #         # and probably became a baseStringClass, which the xmlrpc marshaller
    #         # doesn't know how to handle.
    #         session.postMessages(['javascript', str(result)])


    # This runs before the customizers, because they have to wait for the vm.
    # This means the scriptable event, too.  But also this statement is required,
    # and so far there's nothing that makes anything in enterGame rely on the
    # customization code.
    postJavascript(peer, 'enterGame();' if scriptComplete else
                   'enterGame(complete = true);')


# Rudimentary Access Policy.
TRUSTED_FILENAME = 'etc/trust.python'
class Trust(dict):
    def __init__(self, filename = TRUSTED_FILENAME):
        self.filename = filename
        self.loaded = False

    def load(self):
        try:
            self.update(load_pickle(open(self.filename)))
            self.loaded = True
        except IOError as e:
            from errno import ENOENT
            if e.errno != ENOENT:
                raise

    def save(self):
        save_pickle(self, open(self.filename, 'w'))

    def __contains__(self, avatar):
        not self.loaded and self.load()
        return avatar.idnum == self.get(avatar.name)

    def __iadd__(self, avatar):
        self[avatar.name] = avatar.idnum
        self.save()
        return self

policy = Trust()

# Host Security.
SECURE_DOMAINS = None
TRUST_ALL = False

def getSecureDomains(reload = False):
    global SECURE_DOMAINS, TRUST_ALL
    if SECURE_DOMAINS is None or reload:
        from stuphos import getSection
        securityCfg = getSection('Security')

        # Some builtin domains.
        domains = []
        if isYesValue(securityCfg.get('trust-localhost')):
            # IPv6? Hah!
            try: domains.append(platform.node().lower())
            except: logException(traceback = True)

            domains.append('localhost')
            domains.append('127.0.0.1')

        # Build from config.
        for option in securityCfg.options():
            if option == 'trusted-domain' or \
               option.startswith('trusted-domain.'):
                o = securityCfg.get(option).lower()
                if o == 'all':
                    TRUST_ALL = True
                else:
                    domains.append(o)

        domains = [_f for _f in set(domains) if _f]
        SECURE_DOMAINS = IPAddressGroup(*domains)

    return SECURE_DOMAINS

def isSecureDomain(domainName):
    domains = getSecureDomains()
    if TRUST_ALL:
        return True

    return domainName.lower() in domains

def isFromSecureHost(peer):
    # debugOn()
    if getattr(peer, 'session', None) is not None:
        # Authenticated via SSL.
        if isYesValue(configuration.Security.trust_web_session_adapter):
            return True

    return isSecureDomain(peer.host)

# Communication Constructs.
class PlayerResponse(Exception):
    pass
class DoNotHere(Exception):
    pass

class PlayerAlert(PlayerResponse):
    def __init__(self, alert):
        self.alert = alert
    def deliver(self, peer):
        # Todo: call customized handler.
        print(self.alert, file=peer)

def playerAlert(fmt, *args):
    raise PlayerAlert(fmt % args)

def HandleCommandError(peer, exc = None, full_traceback = True, error_color = 'r', frame_skip = 1):
    if exc is None:
        exc = getSystemException()

    # First, Cascade to MUD.
    HandleException(exc = exc)

    # Then, send to player.
    name = getattr(exc[0], '__name__', '<Unnamed>')
    tb = exc[2]

    # Skip forward frames as long as possible.
    while frame_skip > 0 and tb.tb_next:
        tb = tb.tb_next
        frame_skip -= 1

    # Configure.
    relative = 'relative' if isYesValue(getConfig('traceback-relative')) else True

    # Find (second to?) last frame.
    while tb.tb_next:
        if full_traceback and tb.tb_next:
            print(ShowFrame(tb.tb_frame, name, use_basename = relative), file=peer)

        tb = tb.tb_next

    print('&%s%s&N' % (error_color, ShowFrame(tb.tb_frame, name, exc, use_basename = relative)), file=peer)

    # Whether or not command error was handled -- which it was.
    return True


# Interactive Organization.
from types import ModuleType
# import new

def getPlayerCommands(peer):
    return getSharedCommands()

def getPlayerScope(peer):
    # debugOn()

    # Peer connection states are transient at best, but
    # they can be connected to pre-existing scopes.
    try: return peer.namespace
    except AttributeError: pass

    # Put a persistant namespace module on the avatar,
    # which may exist longer than a network connection.
    # Todo: remember beyond existence of avatar..?
    try: from world import mobile_instance
    except ImportError: pass
    else:
        a = peer.avatar

        if a and isinstance(a, mobile_instance):
            if hasattr(a, 'namespace'):
                return a.namespace

            ns = a.namespace = ModuleType('mud.player.namespace')
            return ns

    # return getSharedScope()

def getSharedCommands():
    from stuphmud.server.player.interfaces import getCommandCenter
    return getCommandCenter(getSharedScope())

def getSharedScope():
    if 'namespace' not in globals():
        global namespace

        namespace = ModuleType('mud.player.namespace')

    return namespace

# Gross -- Scope before Commands?  Decoupled.
# This should just be put in mud.player.interfaces
# (at least i _think_ it's that simple)
shared = getSharedCommands()
ACMD = shared.assignRemoveable # todo: stacked assignment

from stuphos.etc.tools.cmdln import Cmdln, Option
def ACMDLN(verbName, *options, **kwd):
    cmdln = Cmdln(verbName, *options, **kwd)
    if '*' not in verbName:
        verbName += '*'

    def makeCommandHandler(function):
        @ACMD(verbName)
        def doCommand(peer, cmd, argstr):
            try: parsed = cmdln.parseCommand(cmd, argstr)
            except cmdln.HelpExit: return True
            else:
                result = function(peer, parsed)
                if isinstance(result, generator):
                    peer.page('\n'.join(result) + '\n')
                    return True

                return bool(result)

        doCommand.command = function
        return doCommand
    return makeCommandHandler

# todo: ASUBCMD[LN]

def Showing(peer, caption = None, none = None):
    # Use as decorator to generate a paged-string from content and optional caption.
    # Optional 'none' argument shows this as string if no results were obtained.
    def showResults(results):
        def s():
            if caption:
                yield stylizedHeader(caption)

            for r in results:
                yield r

            yield ''

        peer.page('\r\n'.join(s()))

    if none:
        def showOrNone(view):
            results = list(view())
            if results:
                showResults(results)
            else:
                print(none, file=peer)

            return view

        return showOrNone

    def show(view):
        showResults(view())
        return view

    return show



"""CircleMUD mobile database file interpreter.

Version: %(__version__)f

Class: MobileReader
    constructor - first argument is either an open file handle,
        a string path name for a file to open, or a zone number
        for loading from a builtin library path

    parseFile() - initiate process to read entire file.  Calls
        startOfRecords(), newRecord(), recordField(), recordDescription(),
        recordAffection() and endOfRecords().
        Override these methods on subclass to implement your own
        records processing.

    result field - dictionary holding entire parsed-file database

"""

__version__ = 0.1

def usage():return __doc__%{'__version__':__version__}

from .dblib import *

class MobileReader(RecordReader):
    def parseRecord(self):
        """Load record area into logical Python object

        file argument is an open file handle

        Format of textually-defined object mobile:

        #<vnum>
        <Name String (keywords)>~<NL>
        <Short Description>~<NL>
        <Long Room Description (multiline)><NL>
        ~<NL>
        <Description (multiline)><NL>
        ~<NL>
        <ASCII-flag for Mob Bits> <ASCII-flag for Aff Bits> <Alignment> S|E

        Simple Monsters:
        <Level> <Hitroll> <AC> <Hit>d<Mana>+<Move> <DamNoDice>d<DamSizeDice>+<DamRoll>
        <Gold> <Experience>
        <Position> <Default Position> <Gender>

        Circle3 Enhanced Monsters:
        MoveMessage: [0-99]
        BareHandAttack: [0-99]
        Str: [3-25]
        StrAdd: [0-100]
        Int: [3-25]
        Wis: [3-25]
        Dex: [3-25]
        Con: [3-25]
        Cha: [3-25]
        gen_spec_act: <type> <chance> <spell> <self> "<command>"
        SpecProc: <specproc name>
        E
        """

        self.recordField('name', self.tildeString('name'))
        self.recordField('shortdescr', self.tildeString('shortdescr'))
        self.recordField('longdescr', self.tildeString('longdescr'))
        self.recordField('descr', self.tildeString('descr'))

        mobflags, affbits, align, monster_type = self.readLine().split()
        self.recordField('mobflags', ASCIIFlagDecode(mobflags))
        self.recordField('affbits', ASCIIFlagDecode(affbits))
        self.recordField('alignment', align)

        if monster_type in 'SE':
            level, hitroll, ac, hmm, ddd = self.readLine().split()
            self.recordField('level', level)
            self.recordField('hitroll', hitroll)
            self.recordField('armorclass', ac)

            try: hit, mana, move = parseDice(hmm)
            except FormatError as e:
                raise self.error(str(e))

            self.recordField('max_hit', hit)
            self.recordField('max_mana', mana)
            self.recordField('max_move', move)

            try: damnodice, damsizedice, damroll = parseDice(ddd)
            except FormatError as e:
                raise self.error(str(e))

            self.recordField('damnodice', damnodice)
            self.recordField('damsizedice', damsizedice)
            self.recordField('damroll', damroll)

            gold, exp = self.readLine().split()
            self.recordField('gold', gold)
            self.recordField('experience', exp)

            pos, default_pos, gender = self.readLine().split()
            self.recordField('position', pos)
            self.recordField('default_pos', default_pos)
            self.recordField('gender', gender)

        if monster_type == 'E':
            line = self.readLine()
            while line != '':
                code = line.strip()
                if code == '$':
                    # Natural termination of records.
                    return None

                elif code == 'E':
                    # Natural termination of enhanced mobile.
                    break

                colon = line.find(':')
                if colon >= 0:
                    field = line[:colon].strip()
                    value = line[colon+1:]
                    ival = value.strip()
                    if ival.isdigit():
                        value = int(ival)
                else:
                    field = line
                    value = ''

                field = field.lower()

                # Eww.. pollution?
                self.recordField(field, value)

                line = self.readLine()

            if line == '':
                self.EOFNoDollar()

        return self.readNextVnum()



from . import CommandMessage, RegisterCommandClass

class GirlScriptCommandMessage(CommandMessage):
    def Execute(self):
        sourceCode = self.payload.replace('\r', '')

        from stuphos.kernel import Script, Girl, executeGirl
        # from world import heartbeat as vm
        module = Girl(Girl.Module, sourceCode)

        task = Script.Load()
        task.addFrameCall(module)
        return executeGirl(task)

RegisterCommandClass('application/x-stuph-agent-script', GirlScriptCommandMessage)



# Query Language.
try:
    from world import zone as ZoneTable
    from world import room as RoomTable
    from world import mobile as MobileInstanceTable
    from world import item as ItemInstanceTable
    from world import house as HouseTable
    from world import iterate_entities as IterateNativeEntities
    from world import command as CommandType

    NATIVE_TABLE_TYPES = [ZoneTable, RoomTable, MobileInstanceTable, ItemInstanceTable, HouseTable, CommandType]
    world = True

except (ImportError, NameError):
    NATIVE_TABLE_TYPES = []
    world = False

# Object Model.
class base(object):
    # Could be like, a runtime.Object
    @classmethod
    def isSubtype(self, type):
        try: return issubclass(type, self)
        except TypeError:
            return False

    @classmethod
    def isInstanceOf(self, instance):
        try: return self.isSubtype(instance.__class__)
        except AttributeError:
            return False

# Todo: fold native types into object model.
class TableType(base):
    pass

class MobilePrototypeTable(TableType):
    @classmethod
    def iterateRecords(self, callback):
        def iterateZoneMobilePrototypes(zone):
            for mobile in zone.mobiles:
                callback(mobile)

        IterateNativeEntities(ZoneTable, iterateZoneMobilePrototypes)

class ItemPrototypeTable(TableType):
    @classmethod
    def iterateRecords(self, callback):
        def iterateZoneItemPrototypes(zone):
            for item in zone.items:
                callback(item)

        IterateNativeEntities(ZoneTable, iterateZoneItemPrototypes)

class PlayerTable(TableType):
    @classmethod
    def iterateRecords(self, callback):
        pass

# Registered Table Types.
KNOWN_TABLE_TYPES = NATIVE_TABLE_TYPES
if world is not False:
    KNOWN_TABLE_TYPES += [MobilePrototypeTable, ItemPrototypeTable, PlayerTable]

    tableAliases = dict(zones = ZoneTable,
                        rooms = RoomTable,
                        mobiles = MobileInstanceTable,
                        items = ItemInstanceTable,
                        houses = HouseTable,

                        mobile_prototypes = MobilePrototypeTable,
                        item_prototypes = ItemPrototypeTable,
                        players = PlayerTable,

                        command = CommandType)

    tableAliases['prototypes.mobile'] = MobilePrototypeTable
    tableAliases['prototypes.item'] = ItemPrototypeTable
    tableAliases['instances.mobile'] = MobileInstanceTable
    tableAliases['instance.item'] = ItemInstanceTable

def getTable(object):
    if isinstance(object, str):
        return tableAliases.get(object.lower())
    if object in KNOWN_TABLE_TYPES:
        return object

    object = type(object)
    if object in KNOWN_TABLE_TYPES:
        return object

    raise TypeError(object)

def IterateRecords(table, callback):
    if table in NATIVE_TABLE_TYPES:
        return IterateNativeEntities(table, callback)
    if TableType.isSubtype(table):
        return table.iterateRecords(callback)

class Criteria(base, list):
    def __init__(self, *args, **kwd):
        self.extend(args)
        self.extend(iter(kwd.items()))
    def __add__(self, other):
        return Criteria(*(self + other))
    def __getitem__(self, item):
        for (name, value) in list(self.items()):
            if name == name:
                return value

    def __str__(self):
        return '\n'.join('%s: %r' % nv for nv in self)
    def __repr__(self):
        return '%s: %s' % (self.__class__.__name__,
                           ' '.join('%s=%r' % nv for nv in self))

    @classmethod
    def Build(self, **kwd):
        return self(**kwd)

class NullType(object):
    def __repr__(self):
        return 'Null'
    def __str__(self):
        return ''

Null = NullType()

class Query(base):
    def __init__(self, *args, **criteria):
        self.criteria = Criteria(*args, **criteria)
        self.results = []

    def clear(self):
        del self.results[:]
        return self

    # Results:
    def __iter__(self):
        return iter(self.results)

    def __str__(self):
        return str(self.criteria)
    def __repr__(self):
        return '%s: %r' % (self.__class__.__name__, self.criteria)

class TableQuery(Query):
    def _queryTable(self, entity):
        # Todo: and/or
        for (field, clause) in self.criteria:
            try: value = getattr(entity, field)
            except (AttributeError, TypeError, ValueError):
                value = Null # Undefined

            try:
                if clause == value:
                    self.results.append(Result(self, entity, field, value))

            except TypeError:
                pass

    def select(self, *objects):
        # Joins?
        for table in objects:
            if Query.isInstanceOf(table):
                # Subquery.
                assert table is not self
                for result in table:
                    self._queryTable(result.entity)

            else:
                try: table = getTable(table)
                except TypeError: pass
                else: IterateRecords(table, self._queryTable)

        return self

    def fields(self):
        return set(r.field for r in self.results)
    def entities(self):
        return set(r.entity for r in self.results)
    def entityTypes(self):
        return set(type(r.entity) for r in self.results)
    def entitiesByType(self, search_type):
        return [r for r in self.results if type(r.entity) is search_type]
    def fieldsByType(self, search_type):
        return set(r.field for r in self.results if type(r.entity) is search_type)

class Result(base):
    def __init__(self, query, entity, field, value):
        self.query = query
        self.entity = entity
        self.field = field
        self.value = value

    def __repr__(self):
        # Q: should entity repr be r?
        return '%s: %r [%r on %s]' % (self.__class__.__name__,
                                      self.value, self.field,
                                      self.entity)

class Clause(base):
    def __init__(self, this):
        self.this = this
    def __repr__(self):
        return '%s: %r' % (self.__class__.__name__, self.this)

class Equals(Clause):
    def __eq__(self, value):
        return self.this == value

class NotEquals(Equals):
    def __eq__(self, value):
        return self.this != value

class Contains(Clause):
    def __eq__(self, value):
        return self.this in value

class ContainedBy(Clause):
    def __eq__(self, value):
        return value in self.this

class Like(Clause):
    import re
    def __init__(self, pattern):
        self.pattern = self.re.compile(pattern)
    def __eq__(self, value):
        return self.pattern.match(value) is not None
    def __repr__(self):
        return '%s: %s' % (self.__class__.__name__, self.pattern.pattern)

class GreaterThan(Clause):
    def __init__(self, this, OrEqual = False):
        Clause.__init__(self, this)
        self.OrEqual = OrEqual
    def __eq__(self, value):
        if self.OrEqual:
            return self.this >= value
        return self.this > value

class LessThan(Clause):
    def __init__(self, this, OrEqual = False):
        Clause.__init__(self, this)
        self.OrEqual = OrEqual
    def __eq__(self, value):
        if self.OrEqual:
            return self.this <= value
        return self.this < value

class IsTrue(Clause):
    def __eq__(self, value):
        return bool(value)
class IsFalse(Clause):
    def __eq__(self, value):
        return not value

class IsNullType(Clause):
    def __init__(self):
        pass
    def __eq__(self, value):
        return value is Null
class IsNotNullType(Clause):
    def __init__(self):
        pass
    def __eq__(self, value):
        return value is not Null

IsNull = IsNullType()
IsNotNull = IsNotNullType()

class Matches(Clause):
    def __eq__(self, other):
        return self.this(other)

# Aliases.
Q = Query
EQ = Equals
NEQ = NotEquals
Not = NotEquals
NOT = NotEquals
IN = ContainedBy
HAS = Contains
Has = Contains
GT = GreaterThan
LT = LessThan
NULL = IsNull
NOTNULL = IsNotNull
NotNull = IsNotNull

def parseSQL(sql):
    syntax = pysqldap.simpleSQL.parseString(sql)
    if len(syntax) >= 4:
        if syntax[0].lower() != 'select':
            raise SyntaxError('Unknown statement: %r' % syntax[0])
        if syntax[2].lower() != 'from':
            raise SyntaxError('Unknown preposition: %r' % syntax[2])

        columns = syntax[1][0]
        tables = syntax[3]

        if len(syntax) >= 6:
            if syntax[4].lower() != 'where':
                raise SyntaxError('Unknown clause specifier: %r' % syntax[4])

            clause_type = 'and'
            criteria = Criteria()

            for clause in syntax[4]:
                if isinstance(clause, str):
                    if clause.lower() not in ['and', 'or']:
                        raise SyntaxError('Unknown grouping operation: %r' % clause)

                    clause_type = clause

                elif len(clause) is 3:
                    (lh, op, rh) = clause

                    try: clause_handler = CLAUSE_HANDLERS[op]
                    except KeyError:
                        raise SyntaxError('Unknown clause operator: %r' % op)

                    criteria += Criteria(lh = clause_handler(rh))
                    clause_type = None

            if clause_type is not None:
                raise SyntaxError('Expecting clause operator after: %r' % clause_type)

        else:
            # Generate Criteria simply including names that exist.
            criteria = Criteria(*[(field, IsNotNull) for field in columns])

        return (criteria, tables)

    raise SyntaxError('Unknown syntax: %r' % sql)

def showResults(q):
    for entity_type in q.entityTypes():
        yield '%s:' % entity_type

        ten = ' ' * 10
        fields = [(f, len(f) + 10) for f in q.fieldsByType(entity_type)]
        fields = [(f[0], f[1], '%%-%d.%ds' % (f[1], f[1])) for f in fields]
        field_map = dict((fields[nr][0], nr) for nr in range(len(fields)))

        yield ' '.join(f[0] + ten for f in fields)
        yield ' '.join('-' * f[1] for f in fields)

        yield ''

        entity_results = {}
        for result in q.results:
            entity = result.entity
            if type(entity) is entity_type:
                entity_results.setdefault(entity, []).append(result)

        for (entity, results) in entity_results.items():
            # yield '  %s:' % entity

            results = [(field_map[r.field], r) for r in results]
            results.sort()
            results = [r[1] for r in results]

            yield ' '.join(fields[nr][2] % results[nr].value \
                           for nr in range(len(fields)))

        yield ''

# Player Command.
MINIMUM_LEVEL = 115
CLAUSE_HANDLERS = {}

WorldQuery = TableQuery
def doQueryWorld(peer, cmd, argstr):
    # todo: manage queries, views, and reports
    if peer.avatar and peer.avatar.level >= MINIMUM_LEVEL:
        if argstr:
            try: (criteria, tables) = parseSQL(argstr)
            except SyntaxError as e:
                print(e, file=peer)
            else:
                # Perform query.
                q = WorldQuery(*criteria).select(*tables)
                peer.page_string('\r\n'.join(showResults(q)))

        return True

try:
    from stuphmud.server.player import ACMD
    import pysqldap

except ImportError: pass
else: ACMD('mql*')(doQueryWorld)



# Live Deployment of New Code Base.
# Copyright 2021 runphase.com .  All rights reserved
# XXX Uh, dependencies?
from stuphos.system.api import syslog, mudlog

from os.path import join as joinpath, split as splitpath, splitext
from os import unlink as deleteFile, makedirs, system as systemCommand

from errno import EEXIST

import tarfile, gzip
import contextlib
import sys
import pdb

# Platform/Runtime Settings.
RMBIN = 'rm'
DIRUPDATE_BIN = 'dir_update.py'
SVNBIN = 'svn'
DEFAULT_SUBDAEMON_NAME = 'svn-repos'

# Platform Routines.
def shellEscape(arg):
    return repr(arg)

def shellCommand(*args, **kwd):
    cmd = ' '.join(map(shellEscape, args))
    try: stdout = kwd['stdout']
    except KeyError: systemCommand(cmd)
    else:
        try: stderr = kwd['stderr']
        except KeyError:
            systemCommand('%s > %s' % (cmd, shellEscape(stdout)))
        else:
            if stderr == '--':
                systemCommand('%s > %s 2>&1' % (cmd, stdout))
            else:
                systemCommand('%s > %s 2> %s' % (cmd, stdout, stderr))

def deleteDirectory(directory):
    shellCommand(RMBIN, '-Rf', directory)

# Repository subdaemon management.
def getSubDaemonManager():
    ##    from fraun.player.commands.subdaemon import SubDaemonManager
    ##    return SubDaemonManager.get()

    from stuphos.runtime.registry import getObject
    return getObject('SubDaemon::Manager')

class SvnDaemon:
    def __init__(self, name):
        self.name = name

    @contextlib.contextmanager
    def __call__(self, progress_proc):
        manager = getSubDaemonManager()
        assert manager is not None

        daemon = manager.getDaemon(subdaemon_name)
        already_running = daemon.isRunning()

        # todo: with multi-threaded deployment process, we'll need to nest
        # a thread-safe daemon-control locking context here.

        if already_running:
            progress_proc.message('Svn Daemon Already Running')
            yield
        else:
            progress_proc.message('Starting Svn Daemon...')
            daemon.startProcess(())

            try: yield
            finally:
                progress_proc.message('Stopping Svn Daemon...')
                daemon.quitProcess()

@contextlib.contextmanager
def UnmanagedSvnDaemon(progress_proc):
    yield

class SvnReposOp:
    def __init__(self, svnbinpath):
        self.svnbinpath = svnbinpath
    def __call__(self, subcommand, *args, **kwd):
        shellCommand(self.svnbinpath, subcommand, *args, **kwd)

    def checkout(self, xxx_todo_changeme, wc_dir, progress_proc, message_file, quiet = False):
        (repos_url, path) = xxx_todo_changeme
        progress_proc.message('Checking Out Control WC...')
        self('checkout', '%s/%s' % (repos_url, path), wc_dir,
             *(('--quiet',) if quiet else ()),
              stdout = message_file)

        progress_proc.message_file(message_file)

    def commit(self, wc_dir, update_message, progress_proc, message_file):
        progress_proc.message('Committing Changes in Control Directory to Repository...')
        self('commit', wc_dir, '-m', update_message, stdout = message_file)
        progress_proc.message_file(message_file)

    def update(self, active_dir, progress_proc, message_file):
        progress_proc.message('Updating Active Project Directory...')
        self('update', active_dir, stdout = message_file)
        progress_proc.message_file(message_file)

class SvnReposAlreadyCheckedOut(SvnReposOp):
    # Testing... skip actual lengthy checkout.
    def checkout(self, *args, **kwd):
        print('Checkout -- Already Done!')

class WorkingDirProc(object):
    # Manages a temporary working directory in context.

    @contextlib.contextmanager
    def __new__(self, *args, **kwd):
        inst = object.__new__(self)
        inst.__init__(*args, **kwd)

        try: yield inst
        finally: inst.cleanup()

    def __init__(self, tmp_dir_proc, tmp_dir, progress_proc):
        # todo: allow passing base tmp_dir to tmp_dir_proc ala mkdtemp/dir = ??
        self.tmp_dir_base = tmp_dir_proc() if tmp_dir is None else tmp_dir
        self.progress_proc = progress_proc

    def cleanup(self):
        self.progress_proc.message('Cleaning up temporary %r...' % self.tmp_dir_base)
        deleteDirectory(self.tmp_dir_base)

    # Todo: keep track of directories for a checked newTmpXX
    def newTmpDir(self, *key):
        path = joinpath(self.tmp_dir_base, *key)
        self.ensureDirectories(path)
        return path

    def newTmpFile(self, *key):
        return joinpath(self.tmp_dir_base, *key)

    def ensureDirectories(self, filename):
        # This creates the directories leading up to the base.
        (path, fn) = splitpath(filename)
        if not fn:
            raise ValueError('Directory path has no base filename! (%s)' % filename)

        try: makedirs(path)
        except OSError as e:
            if e.args[0] != EEXIST:
                raise

def DirUpdate(wd, src, dest, progress_proc, message_file):
    # This assumes that the dir_update.py script is already installed on the system/user path.
    progress_proc.message('Running Directory Differential...')
    update_script = wd.newTmpDir('update_script.sh')
    shellCommand(DIRUPDATE_BIN, dest, src, '-o', update_script, stdout = message_file)
    progress_proc.message_file(message_file)
    shellCommand('source', update_script, stdout = message_file)
    progress_proc.message_file(message_file)

def ExtractArchive(archive, wd, progress_proc, message_file):
    progress_proc.message('Extracting Archive %s...' % archive)
    archive.extractToDirectory(wd, message_file)
    progress_proc.message_file(message_file)

def TempDir():
    from tempfile import mkdtemp
    return mkdtemp(dir = '/tmp', prefix = 'stuphdeploy-')

def GenerateUpdateMessage(svn_repos_proc, control):
    return ''
    return svn_repos_proc.log(control) # from last deployment?

class FileArchive:
    def __init__(self, filename, archive_name):
        self.filename = filename
        self.archive_name = archive_name

    def __str__(self):
        return self.archive_name

    def getExtractor(self):
        filename = self.filename
        if filename.endswith('.tgz') or filename.endswith('.tar.gz'):
            return tarfile.TarFile(fileobj = gzip.GzipFile(filename)).extractall
        elif filename.endswith('.tar.bz2'):
            import bz2
            return tarfile.TarFile(fileobj = bz2.BZ2File(filename)).extractall
        elif filename.endswith('.zip'):
            import zipfile
            return zipfile.ZipFile(filename).extractall

        raise TypeError('Unknown archive type: %s' % filename)

    def extractToDirectory(self, directory, message_file):
        # Do this verbosely?
        extractor = self.getExtractor()
        extractor(directory)

@apply
class MudlogProgress:
    def message(self, message):
        mudlog(message)
    def message_file(self, message_file):
        pass

    def start(self, archive):
        syslog('SYSTEM: Deployment Started: %s' % archive)
    def exception(self, etype, value, tb):
        mudlog('SYSTEM: Deployment Failure: %s' % value, file = True)
    def success(self):
        mudlog('SYSTEM: Deployment Complete.', file = True)

class ProgressStream:
    def __init__(self, stream, synchronize):
        self.stream = stream
        if synchronize:
            self.synchronize = self.asynchronousCall
        else:
            self.synchronize = self.synchronousCall

    def asynchronousCall(self, *args, **kwd):
        from stuphos import enqueueHeartbeatTask
        enqueueHeartbeatTask(*args, **kwd)

    def synchronousCall(self, func, *args, **kwd):
        func(*args, **kwd)

    def getStream(self):
        return self.stream
    def sendMessage(self, message):
        self.getStream().write(message)

    EOL = '\n'
    def message(self, message):
        self.synchronize(self.sendMessage, message + self.EOL)

    PAGELENGTH = 5
    def message_file(self, message_file):
        try: fp = open(message_file)
        except IOError: pass
        else:
            output = []
            nr = 0
            for line in fp:
                output.append(line.rstrip())

                nr += 1
                if nr > self.PAGELENGTH:
                    output.append('(more ...)')
                    break

            if output:
                self.synchronize(self.sendMessage, self.EOL.join(output) + self.EOLF)

    def start(self, archive):
        self.synchronize(syslog, 'SYSTEM: Deployment Started: %s' % archive)
    def exception(self, etype, value, tb):
        self.synchronize(mudlog, 'SYSTEM: Deployment Failure: %s' % value)
        self.synchronize(syslog, 'SYSTEM: Deployment Failure: %s' % value)
    def success(self):
        self.synchronize(mudlog, 'SYSTEM: Deployment Complete.')
        self.synchronize(syslog, 'SYSTEM: Deployment Complete.')

class UploadedUpdateMessage:
    def __init__(self, update_message):
        self.update_message = update_message
    def __call__(self, svn_repos_proc, control_wd):
        return self.update_message

class DeploymentConfig:
    @classmethod
    def Load(self, name):
        from stuphos import getConfigObject
        cfg = getConfigObject()['Deployment']

        assert cfg is not None
        cfgfile = cfg.get(name)
        if cfgfile is not None:
            from stuphos.etc.config import loadConfig
            return DeploymentConfig(name, loadConfig(cfgfile)['Deployment'])

        from configparser import NoOptionError
        raise NoOptionError(name, 'Deployment')

    def __init__(self, name, cfg):
        self.key = cfg.get('key')

        self.proto = cfg.get('proto')
        self.host = cfg.get('host')
        self.port = cfg.get('port')

        self.subpath = cfg.get('subpath') or ''
        self.active_dir = cfg.get('active-dir')

        self.subdaemon_name = cfg.get('subdaemon-name')

        assert self.key is not None
        assert self.proto is not None
        assert self.host is not None
        assert self.active_dir is not None

def multithreadOn(state, function, *args, **kwd):
    if not state:
        return function(*args, **kwd)

    from _thread import start_new_thread as nth
    nth(function, args, kwd)

# Main Process.
def LiveDeployment(xxx_todo_changeme1, archive, active_dir,
                   tmp_dir = None, tmp_dir_proc = TempDir,
                   extract_archive_proc = ExtractArchive,
                   working_dir_proc = WorkingDirProc,
                   update_message_proc = GenerateUpdateMessage,
                   svn_repos_proc = SvnReposOp(SVNBIN),
                   dir_update_proc = DirUpdate,
                   svn_daemon_proc = UnmanagedSvnDaemon,
                   progress_proc = MudlogProgress):

    (repos_url, repos_path, subpath) = xxx_todo_changeme1
    progress_proc.start(archive)
    try:
        with working_dir_proc(tmp_dir_proc, tmp_dir, progress_proc) as wd:
            with svn_daemon_proc(progress_proc):
                control_wd = wd.newTmpDir('control')
                archive_wd = wd.newTmpDir('archive')
                message_file = wd.newTmpFile('message_file.txt')

                # Extract deployment archive and check out the directory from the control repository.
                extract_archive_proc(archive, archive_wd, progress_proc, message_file)
                svn_repos_proc.checkout((repos_url, subpath), control_wd, progress_proc, message_file)

                # Perform differential directory upgrade, acquiring commit message.
                dir_update_proc(wd, archive_wd, control_wd, progress_proc, message_file)
                update_message = update_message_proc(svn_repos_proc, control_wd)

                # Commit dir-update to control directory and then update the active filesystem.
                svn_repos_proc.commit(control_wd, update_message, progress_proc, message_file)
                svn_repos_proc.update(active_dir, progress_proc, message_file)
    except:
        progress_proc.exception(*sys.exc_info())
        raise
    else:
        progress_proc.success()

    #   Keep track of already-done deployments and their md5sums...

# Varieties:
def LiveDeploymentLocal(repos_path, archive, active_dir, *args, **kwd):
    return LiveDeployment(('file://%s' % repos_path, repos_path, ''),
                          archive, active_dir, *args, **kwd)

def LiveDeploymentUrl(proto, xxx_todo_changeme2, subpath, archive, active_dir, *args, **kwd):
    (host, port) = xxx_todo_changeme2
    hostport = host if port is None else '%s:%d' % (host, port)
    return LiveDeployment(('%s://%s' % (proto, hostport), '', subpath),
                          archive, active_dir, *args, **kwd)

def LiveDeploymentUrlFromConfigAndArchive(config, archive, **kwd):
    # Resolve config object.
    if isinstance(config, str):
        config = DeploymentConfig.Load(config) # as name
    assert isinstance(config, DeploymentConfig)

    # Load configuration into arguments.
    args = (config.proto, (config.host, config.port),
            config.subpath, archive, config.active_dir)

    # Configure ancillary options.
    updateMessage = kwd.pop('updateMessage')
    if updateMessage:
        kwd['update_message_proc'] = UploadedUpdateMessage(updateMessage)

    if config.subdaemon_name:
        kwd['svn_subdaemon_proc'] = SvnDaemonObject(config.subdaemon_name)

    # Dispatch into Compartment.
    return LiveDeploymentUrl(*args, **kwd)

def LiveDeploymentUrlFromConfigAndArchiveOnMultithread(multithread, *args, **kwd):
    return multithreadOn(multithread, LiveDeploymentUrlFromConfigAndArchive, *args, **kwd)


# Object-Oriented -- todo: fold everything under this class (so it can be overrided easy).
##    class Deployment:
##        def __init__(self, config):
##            self.config = config
##        def liveDeploy(self):
##            pass

@runtime.api('MUD::Deployment::API')
class DeploymentAPI:
    LiveDeployment = staticmethod(LiveDeployment)
    LiveDeploymentUrl = staticmethod(LiveDeploymentUrl)
    LiveDeploymentLocal = staticmethod(LiveDeploymentLocal)

    LiveDeploymentUrlFromConfigAndArchiveOnMultithread = \
        staticmethod(LiveDeploymentUrlFromConfigAndArchiveOnMultithread)

    LoadDeploymentConfig = DeploymentConfig.Load

    getUploadedUpdateMessage = UploadedUpdateMessage
    getSvnDaemonObject = SvnDaemon

    @classmethod
    def getMudlogProgressObject(self):
        return MudlogProgress

    @classmethod
    def getArchiveBaseClass(self):
        return FileArchive

    @classmethod
    def getProgressStreamBaseClass(self):
        return ProgressStream



# Required for Command Dispatch:
from stuphmud.server.player import ACMD, HandleCommandError, PlayerAlert, playerAlert
from stuphos.etc.tools import getSystemException, stuphColorFormat
from stuphos.runtime.registry import getObject, registerObject, ObjectAlreadyRegistered

# Control-Point Support:
from stuphos.etc.tools import clearLineCache

from stuphmud.server.zones.specials import LookupSpecialObject
from stuphos import getConfig

import world
from world.unstructured import setWeather
import imp

def apply(function, *args, **kwd):
    return function(*args, **kwd)

def reloadModule(name):
    return imp.reload(__import__(name, globals(), locals(), ['']))

def parameters(min, max = None, doc = None):
    def makeParameterized(function):
        function.nParameters = (min, max)
        function.wizDoc = doc
        return function
    return makeParameterized

def paramArgstr(function):
    function.paramsArgstr = True
    return function

def parameterized_method(min, max = None, doc = None):
    def makeParameterized(function):
        def parameterizedMethod(self, *args, **kwd):
            return function(*args, **kwd)

        try: parameterizedMethod.__name__ = function.__name__
        except AttributeError: pass

        parameterizedMethod.nParameters = (min, max)
        parameterizedMethod.wizDoc = doc
        return parameterizedMethod
    return makeParameterized

def uncontrollable(function):
    function._uncontrollable = True
    return function

@apply
class ControlPoints:
    class Cancel(Exception):
        pass

    @parameters(1, 3)
    def set(self, peer, feature, attribute = None, value = None):
        feature = feature.lower()
        if feature == 'weather':
            attribute = attribute.lower() if attribute else ''
            if attribute == 'pressure':
                value = max(960, min(1040, int(value)))
                setWeather('pressure', value)
                playerAlert('Weather pressure set to: %d' % value)

            elif attribute == 'sky':
                playerAlert('Not yet implemented!')
            else:
                playerAlert('Unknown weather-set option: %r' % attribute)

        elif not feature:
            raise self.Cancel
        else:
            playerAlert('Unknown set feature: %r' % feature)

    # Should be part of system:
    @parameters(None)
    def clear_cache(self, peer):
        clearLineCache()
        playerAlert('Line-cache cleared!')

    @parameters(0)
    def load(self, peer, *modules):
        # Put them in the main namespace.
        import __main__ as main
        main = main.__dict__

        for m in modules:
            name = m.split('.')[0]
            try: main[name] = __import__(m)
            except ImportError:
                HandleCommandError(peer, getSystemException())
            else:
                print('Loaded: %s' % m)

    @parameters(0)
    def reload(self, peer, *modules):
        if not modules:
            modules = [__name__]

        modules = list(map(reloadModule, modules))
        clearLineCache()
        playerAlert('RELOADED:\r\n%s\r\n' % '\r\n'.join(map(str, modules)))

    @parameters(1)
    def reload_package(self, peer, package):
        # Delete all entries from the system modules that are under the
        # given package.  If the package-name is malformed, who cares.
        unloaded = self._unload_package(package)
        if unloaded:
            print('Unloaded:\n    %s' % '\n    '.join(unloaded), file=peer)
            print(file=peer)
        else:
            print('No submodules unloaded.', file=peer)

        # Then import the top-level.
        print('Reloading %s...' % package, file=peer)
        __import__(package)

    @parameters(0)
    def unload(self, peer, *packages):
        for p in packages:
            print('Unloaded:\n    %s' % \
                  '\n    '.join(self._unload_package(p)), file=peer)

    @uncontrollable
    def _unload_package(self, package):
        pGroup = package.split('.')
        pLength = len(pGroup)
        assert package # Not necessary, but we're paranoid.
        unloaded = []

        def isSubmodule(m):
            return m.split('.')[:pLength] == pGroup

        from sys import modules
        for m in filter(isSubmodule, iter(modules.keys())):
            unloaded.append(m)
            del modules[m]

        return unloaded

    @parameters(3, doc = 'Usage: assign {room|item|mobile} <vnum> <fully.qualified.module.Special>')
    def assign(self, peer, spec_type, vnum, fqName):
        player = peer.avatar
        if player is None:
            playerAlert('You must be connected to a player to use this command.')

        spec_type = spec_type.lower()
        if spec_type == 'room':
            if vnum == 'here':
                room = player.room
            elif vnum.isdigit():
                room = world.room(int(vnum))
            else:
                playerAlert('Usage: assign room <vnum | here> <name>')

            if room is None:
                playerAlert('Room not found: %r' % vnum)

            proc = LookupSpecialObject(fqName, spec_type = spec_type)
            if proc is not None:
                from stuphmud.server.zones.specials import SpecialManager
                SpecialManager.AssignRoom(room, proc)
                playerAlert('Installed %r into %r.' % (proc, room))
            else:
                playerAlert('Unknown procedure: %r' % fqName)

        elif spec_type == 'mobile':
            mobile = world.mobile(int(vnum)) if vnum.isdigit() else player.find(vnum)
            if mobile is None or not isinstance(mobile, world.mobile):
                playerAlert('Mobile not found: %r' % vnum)

            proc = LookupSpecialObject(fqName, spec_type = spec_type)
            if proc is not None:
                from stuphmud.server.zones.specials import SpecialManager
                SpecialManager.AssignMobile(mobile, proc)
                playerAlert('Installed %r into %r.' % (proc, mobile))
            else:
                playerAlert('Unknown procedure: %r' % fqName)

        elif spec_type == 'item':
            item = world.item(int(vnum)) if vnum.isdigit() else player.find(vnum)
            if item is None or not isinstance(item, world.item):
                playerAlert('Item not found: %r' % vnum)

            proc = LookupSpecialObject(fqName, spec_type = spec_type)
            if proc is not None:
                from stuphmud.server.zones.specials import SpecialManager
                SpecialManager.AssignMobile(item, proc)
                playerAlert('Installed %r into %r.' % (proc, item))
            else:
                playerAlert('Unknown procedure: %r' % fqName)

        else:
            playerAlert('Unknown special type: %r' % spec_type)

    @parameters(0, 2)
    def registry(self, peer, action = None, name = None):
        if action == 'delete':
            assert name
            from stuphos.runtime.registry import delObject
            delObject(name)

        elif action in (None, 'show'):
            from stuphos.runtime.registry import getRegistry
            from stuphos.etc.tools import maxWidth
            reg = getRegistry() # runtime_registry
            items = iter(reg.items())

            max_length = maxWidth((n for (n, v) in items), 10)
            fmt = '%%-%ds -- %%.%ds' % (max_length, max(40, 80 - max_length - 4))

            def escapeObjectRepr(v):
                return str(v).replace('\n', '\\n') # todo: use encode('string-escape')??

            # Sort ascending.
            items = list(reg.items())
            items.sort(lambda a, b:cmp(a[0], b[0]))
            peer.page_string('\r\n'.join(fmt % (n, escapeObjectRepr(v)) \
                                         for (n, v) in items) + '\r\n')

    def remote_debug(self, peer, command):
        if command == 'break':
            from rpdb import Rdb
            Rdb().set_trace()

    def charm(self, peer):
        # -system-path register /cygdrive/c/Program+Files/JetBrains/PyCharm+1.1.1/helpers
        from pydev import pydevd
        pydevd.settrace('localhost', port=4448, stdoutToServer=True, stderrToServer=True)

    try:
        from .web_debug import doBreakOnUrl
        break_url = staticmethod(doBreakOnUrl)
    except ImportError:
        pass

    from stuphmud.server.player.interfaces.code import doManageShellEvaluators
    shell = staticmethod(doManageShellEvaluators)

    # todo: delegate more dynamically.
    # from .olc import doCompareWorld as compare_world
    # compare_world = staticmethod(compare_world)

    # from .xdebug import doXDebug as xdebug
    # xdebug = staticmethod(xdebug)

    # from .system import doSystemPathControl, doSystemModuleControl, doSystemCheckpointControl
    # system_path = staticmethod(doSystemPathControl)
    # system_modules = staticmethod(doSystemModuleControl)
    # system_checkpoint = staticmethod(doSystemCheckpointControl)

    # from .profiling import doProfileHeartbeat
    # profile_heartbeat = staticmethod(doProfileHeartbeat)

    @parameters(1)
    def view_code(self, peer, moduleName):
        # Or, lookup only modules/submodules in system already.
        m = __import__(moduleName, fromlist = ['']) # LookupObject?
        try: f = m.__file__
        except AttributeError: pass
        else:
            if f[-4:].lower() == '.pyc':
                f = f[:-1]
            if f[-3:].lower() == '.py':
                code = open(f).read()
                code = stuphColorFormat(code)

                peer.page_string(code)

    @parameters(1, 3)
    def upload_zone(self, peer, vnum, database = 'primary', vnum2 = None):
        # Just transfer zone entry info from programmed archives setting to specified database.
        # STUPHLIB -> ORM
        showAvailable = False
        showDetails = False
        if vnum in ['--available', '--index']:
            showAvailable = True
        elif vnum in ['--details']:
            showDetails = True
        else:
            vnum = int(vnum)

        from stuphlib import wldlib, dblib
        world_path = getConfig('world-path', 'MUD')
        base = dblib.LocalIndex(world_path)

        w = wldlib.World(base)
        w.loadFromZoneIndex(cascade = False)

        if showDetails:
            if database == '--mobile':
                vnum2 = int(vnum2)
                zone = w.loadZone(vnum2 // 100)

                for mob in zone.mobiles:
                    if mob.vnum == vnum2:
                        def report():
                            yield '[#%8d] %s' % (mob.vnum, mob.shortdescr)
                            yield 'Keywords: &y%s&N' % mob.name
                            yield 'Long Description:'
                            yield '&y%s&N' % mob.longdescr
                            yield 'Description:'
                            yield '&y%s&N' % mob.descr
                            yield 'Flags: &y%s&N Align: &y%s&N Level: &y%s&N Armor: &y%s&N' % \
                                  (mob.mobflags, mob.alignment, mob.level, mob.armorclass)
                            yield 'Max Hit: &y%5s&N Max Move: &y%5s&N Max Mana: &y%5s&N Hitroll: &y%s&N Damroll: &y%s&N' % \
                                  (mob.max_hit, mob.max_move, mob.max_mana, mob.hitroll, mob.damroll)
                            yield 'Damage Dice: &y%s&N Size Damage Dice: &y%s&N Gold: &y%s&N Exp: &y%s&N' % \
                                  (mob.damnodice, mob.damsizedice, mob.gold, mob.experience)
                            # position = int(mob.position), default_position = int(mob.default_pos),
                            yield 'Sex: 0 Walk Type: &y%s&N' % getattr(mob, 'movemessage', 0)
                            yield ''

                        peer.page_string('\r\n'.join(report()))
                        break

                return

            if database == '--room':
                vnum2 = int(vnum2)
                zone = w.loadZone(vnum2 // 100)

                for room in zone.rooms:
                    if room.vnum == vnum2:
                        def report():
                            yield '[#%8d] %s [&y%s&N/&y%s&N]' % (room.vnum, room.name, room.flags, room.sector)
                            yield '&y%s&N' % room.descr
                            yield ''

                            for (dirCode, exit) in room.exits.items():
                                yield '    %s -- #&y%s&N [&y%s&N/&Y%s&N]' % (dirCode, exit['room-link'],
                                                                             exit['flags'], exit['key'])
                                yield '    Keywords: &y%s&N' % exit['keyword']
                                yield '    &y%s&N' % exit['descr']

                            if room.exits:
                                yield ''

                        peer.page_string('\r\n'.join(report()))
                        break

                return

            database = int(database)
            zone = w.loadZone(database)

            def report():
                yield '%r:' % zone
                yield ''

                yield 'Rooms: vnum name #exits flags'
                for room in zone.rooms:
                    yield '    [#%8d] %-40.40s %4d %s' % (room.vnum, room.name,
                                                          len(room.exits),
                                                          room.flags)
                yield ''

                yield 'Items: vnum shortdescr type'
                for item in zone.items:
                    yield '    [#%8d] %-40.40s %-2d' % (item.vnum, item.shortdescr, item.type)
                yield ''

                yield 'Mobiles: vnum shortdescr level max_hit'
                for mob in zone.mobiles:
                    yield '    [#%8d] %-40.40s %-3s %-5s' % (mob.vnum, mob.shortdescr, mob.level, mob.max_hit)
                yield ''

            peer.page_string('\r\n'.join(report()))
            return

        #start = time.now()
        if showAvailable:
            def report():
                #loadTime = time.now() - start
                for zone in sorted(iter(w.zoneMap.items()), key = lambda z: z[0]):
                    # todo: filter out already-loaded zones.
                    yield '[#%-8d] %s' % (zone[0], zone[1].name)

                yield ''

            peer.page_string('\r\n'.join(report()))
            return

        zone = w.loadZone(vnum)
        # zone = w.lookupZone(vnum)

        # Calculate the real number (*!@#$) using the current world as index.
        # XXX This won't work for sequentially uploaded zones! (Just once)
        index = [z.vnum for z in world.iterent(world.zone)]
        index.append(vnum)
        index.sort() # ascending
        rnum = index.index(vnum)

        # Upload zone, room, object and mobile prototype info.
        # Todo: assert distinct
        from stuphos.management import db, orm
        import sqlobject.dberrors

        # Todo: Do in transaction.
        with db.dbCore.hubThread(database) as conn:
            z = orm.Zones(id = vnum,
                          name = zone.name, lifespan = zone.lifespan, age = 0,
                          bottom = 0, # zone.bottom,
                          top = zone.top, flags = zone.flags,
                          reset_mode = getattr(zone, 'reset-mode', 0),
                          continent = zone.continent)

            z.sync()

            for (i, (command, if_flag, args)) in enumerate(zone.commands):
                args = list(map(int, args))
                c = orm.ZoneCommands(znum = i, command = c, if_flag = if_flag,
                                     arg1 = args[0], arg2 = args[1], arg3 = args[2],
                                     line = '')

                c.sync()

            # XXX This demands a shift up for all zone rnums in the rooms table above this new zone.
            # update rooms set zone = zone + 1 where zone >= $rnum
            # using conn
            u = conn.getConnection().cursor()
            u.execute('update rooms set zone = zone + 1 where zone >= %d' % rnum)

            try:
                for room in zone.rooms:
                    r = orm.Rooms(zone = rnum, name = room.name, sector_type = room.sector,
                                  description = room.descr, flags = room.flags,
                                  light = 0, number = room.vnum)
                    r.sync()

                    # Todo: load extra descriptions.
            except sqlobject.dberrors.DuplicateEntryError:
                pass
            except Exception as e:
                print(e.__class__.__module__, e.__class__.__name__, e, file=peer)

            try:
                for room in zone.rooms:
                    for (dirCode, exit) in room.exits.items():
                        e = orm.RoomDirections(rnum = room.vnum, direction = dirCode,
                                               description = exit['descr'],
                                               keyword = exit['keyword'],
                                               exit_flags = exit['flags'],
                                               exit_key = exit['key'],
                                               exit_destination = exit['room-link'])
                        e.sync()
            except sqlobject.dberrors.DuplicateEntryError:
                pass
            except Exception as e:
                print(e.__class__.__module__, e.__class__.__name__, e, file=peer)

            try:
                for item in zone.items:
                    i = orm.ObjectPrototypes(number = item.vnum,
                                             flags = item.extraflags, name = item.name,
                                             description = item.longdescr,
                                             short_description = item.shortdescr,
                                             action_description = item.actiondescr,
                                             value1 = item.value1,
                                             value2 = item.value2,
                                             value3 = item.value3,
                                             value4 = item.value4,
                                             type = item.type,
                                             wear_flags = item.wearflags,
                                             extra_flags = 0, # item.extra_flags,
                                             anti_flags = item.antiflags,
                                             weight = item.weight,
                                             cost = item.cost,
                                             cost_per_day = 0, # item.cost_per_day,
                                             timer = getattr(item, 'timer', 0),
                                             trap = 0, # item.trap,
                                             bitvector = 0,
                                             spec_proc = '')

                    i.sync()
            except sqlobject.dberrors.DuplicateEntryError:
                pass
            except Exception as e:
                print(e.__class__.__module__, e.__class__.__name__, e, file=peer)

            try:
                for mob in zone.mobiles:
                    m = orm.MobPrototypes(id = mob.vnum, name = mob.name, short_desc = mob.shortdescr,
                                          long_desc = mob.longdescr, description = mob.descr,
                                          flags = mob.mobflags, affected_by = mob.affbits,
                                          alignment = int(mob.alignment),
                                          a_str = 11, # mob.strength,
                                          a_str_add = 11, # mob.strength_addition
                                          a_int = 11, # mob.intelligence
                                          a_wis = 11, # mob.wisdom
                                          a_dex = 11, # mob.dexterity
                                          a_con = 11, # mob.constitution
                                          a_cha = 11, # mob.charisma
                                          level = int(mob.level), hitroll = int(mob.hitroll),
                                          armor = int(mob.armorclass), max_hit = int(mob.max_hit),
                                          hit = 0, # mob.hit,
                                          max_move = int(mob.max_move),
                                          move = 0, # mob.move
                                          max_mana = int(mob.max_mana),
                                          mana = 0, # mob.mana
                                          damage_dice = int(mob.damnodice), damage_dice_faces = int(mob.damsizedice),
                                          damroll = int(mob.damroll), gold = int(mob.gold), exp = int(mob.experience),
                                          position = int(mob.position), default_position = int(mob.default_pos),
                                          sex = 0, # mob.sex
                                          chclass = 0, weight = 0, height = 0,
                                          walk_type = getattr(mob, 'movemessage', 0),
                                          attack_type = 0, # mob.attack_type,
                                          spec_proc = getattr(mob, 'specproc', ''))
                    m.sync()

            except sqlobject.dberrors.DuplicateEntryError:
                pass
            except Exception as e:
                print(e.__class__.__module__, e.__class__.__name__, e, file=peer)

            # todo: shops

            print('STUPHLIB[%s] -> ORM[%s]:' % (world_path, database), file=peer)
            print(repr(zone), file=peer)

    @parameters(1, 2)
    def delete_zone(self, peer, vnum, database = 'primary'):
        # Nuke an entry in the database.
        vnum = int(vnum)

        from stuphos.management import db, orm
        with db.dbCore.hubThread(database) as conn:
            for zone in orm.Zones.selectBy(id = vnum):
                print('Destroying', zone, file=peer)
                zone.destroySelf()

                # todo: destroy rooms, zone commands, object and mobile prototypes, shops.
                # delete from object_prototypes where number / 100 = 20

            # todo: decrease all zone rnums in Rooms table over this zone, using conn.
            u = conn.getConnection().cursor()
            #u.execute('update rooms set zone = zone - 1 where zone > %d' % rnum)

    @parameters(0, 1)
    def show_players(self, peer, database = 'primary'):
        from stuphos.management import db, orm
        from io import StringIO
        import csv

        COLFMT = '#%8s %-20.20s %-30.30s %-40.40s %3s %8s %10s %8s'

        with db.dbCore.hubThread(database) as conn:
            u = conn.getConnection().cursor()
            u.execute('select (id, name, email, host, level, gold, bank_gold, played) from players order by host')

            def report():
                yield COLFMT % ('ID #', 'Name', 'Email', 'Host', 'Lvl', 'Gold', 'Bank', 'Played')

                for row in u:
                    # Parse each row.
                    record = row[0]
                    # '(1144,Cal,calvindastud@hotmail.com,"",105,9979704,1050000000,1698413)'
                    if record[0] == '(' and record[-1] == ')':
                        record = record[1:-1]

                    ((id, name, email, host, level, gold, bank, played),) = csv.reader(StringIO(record))

                    yield COLFMT % (id, name, email, host, level, gold, bank, played)

                yield ''

            text = '\r\n'.join(report())

        peer.page_string(text)

def getMinimumLevel(cmd):
    return 115 # cmd.level

#@runtime.api
@apply
class API:
    NAME = 'Wizard::Control::API'
    REGISTRY = NAME + '::Registry'

    def registerControlPoint(self, name, function):
        self.getRegistry()[name] = function
    def unregisterControlPoint(self, name):
        del self.getRegistry()[name]

    def getRegistry(self):
        return getObject(self.REGISTRY, create = dict)
    def getRegistryObject(self):
        return self.RegistryObject(API.getRegistry())

    def register(self, name): # parameter_info
        # Decorator-Maker:
        def registerFunction(function):
            self.registerControlPoint(name, function)
            return function

        return registerFunction

    class RegistryObject:
        # HAck -- todo, unify this more.
        def __init__(self, d):
            self.__dict__ = d

    parameters = staticmethod(parameters)
    paramArgstr = staticmethod(paramArgstr)

try: registerObject(API.NAME, API)
except ObjectAlreadyRegistered:
    pass

@apply
class WizControllers:
    def __iter__(self):
        yield ControlPoints # Builtin
        yield API.getRegistryObject()

@ACMD('-')
def doWizControl(peer, cmd, argstr):
    if peer.avatar is not None and peer.avatar.level >= getMinimumLevel(cmd):
        args = argstr.split() if argstr else []
        if args:
            larg = args[0]
            if not larg.startswith('_'):
                i = larg.find(':')
                if i < 0:
                    name = larg
                    args = tuple(args[1:])
                else:
                    name = larg[:i]
                    args = (larg[i+1:],) + tuple(args[1:])

                name = name.lower().replace('-', '_')
                for wizctlr in WizControllers:
                    ctlpt = getattr(wizctlr, name, None)
                    if getattr(ctlpt, '_uncontrollable', False):
                        continue

                    try: params = ctlpt.nParameters
                    except AttributeError: pass
                    else:
                        n = len(args)
                        try:
                            if params[0] is None:
                                if n:
                                    playerAlert('%r takes no parameters.' % name)
                            elif n < params[0]:
                                playerAlert('%r takes at least %d parameters.' % (name, params[0]))
                            elif params[1] is not None and n > params[1]:
                                playerAlert('%r takes at most %d parameters.' % (name, params[1]))

                        except PlayerAlert as e:
                            doc = getattr(ctlpt, 'wizDoc', None)
                            if doc:
                                print(doc, file=peer)

                            e.deliver(peer)
                            return True

                    try: params = ctlpt.paramsArgstr
                    except AttributeError: pass
                    else:
                        if params:
                            args = [argstr]

                    if callable(ctlpt):
                        try: ctlpt(peer, *args)
                        except ControlPoints.Cancel: return False
                        except PlayerAlert as e: e.deliver(peer)
                        except: HandleCommandError(peer, full_traceback = False)

                        return True

# Command-Programming:
##    from mud.player import ACMDLN, Option, isFromSecureHost
##    @ACMDLN('.', Option('--verb-name'),
##                 Option('--simple', action = 'store_true'),
##                 Option('--structured', action = 'store_false', dest = 'simple'),
##                 Option('--notebook-page'))
##    def doProgramVerb(player, command):
##        # De/Generate command code.
##        if player.avatar is not None and player.avatar.supreme and isFromSecureHost(player):
##            subCmd = command.nextArg().lower()
##            if subCmd == 'program':
##                # Decide verb name.
##                verbCode = command.options.verb_name or command.nextArg()
##                if not verbCode:
##                    print >> player, command.help()
##                    return True
##
##                # Decide command type (simple or structured)
##                # No: always complex, currently (because of api)
##                ##    if command.options.simple:
##                ##        from mud.player import ACMD as newVerb
##                ##        parameters = 'player, '
##                ##    else:
##                ##        from mud.player import ACMDLN as newVerb
##                ##        parameters = 'player, command'
##
##                # Invoke lambda-programming api.
##                from mud.player.commands import programVerbCommand, BuiltCommandError
##
##                # Todo: allow specification of cmdln options, as prologue:
##                #
##                # [option-name]
##                # short: o
##                # type: str
##                # default: not much
##                # action: store
##                # dest: option
##                #
##                # [another-option]
##                # action: store-true
##                #
##                # if not command.options.another_option:
##                #    print >> player, command.options.option
##                #
##
##                name = command.options.notebook_page
##                if name:
##                    # Interface with notebook api.
##                    # nb = runtime.Player.Notebook.API.Open(player)
##                    nb = runtime[runtime.Player.Notebook.API.Open](player)
##
##                    try: page = nb.getPageContent(name)
##                    except KeyError:
##                        print >> player, 'No page named: %r' % name
##                    else:
##                        # Build and register command from notebook page.
##                        try: programVerbCommand(verbCode, 'python', page)
##                        except BuiltCommandError, build:
##                            print >> player, 'Declaring Code:\n    %s\n' % \
##                                     '\n    '.join(build.declaration.split('\n'))
##
##                            from mud.player import HandleCommandError
##                            HandleCommandError(player, (build.syntax.__class__,
##                                                        build.syntax,
##                                                        build.tb))
##                        else:
##                            # New command registered.
##                            print >> player, 'Built %r from notebook: %r' % (verbCode, name)
##
##                    return True
##
##                # Start interactive programming!.
##                from mud.player.editor import Edit
##                @Edit(player, '')
##                def programVerb(peer, content):
##                    # Build and compile function definition.
##                    if not content.strip():
##                        print >> peer, 'Programming aborted.'
##                    else:
##                        try: programVerbCommand(verb, 'python', content)
##                        except BuiltCommandError, build:
##                            print >> peer, 'Declaring Code:\n    %s\n' % \
##                                  '\n    '.join(build.declaration.split('\n'))
##
##                            from mud.player import HandleCommandError
##                            HandleCommandError(peer, (build.syntax.__class__,
##                                                      build.syntax,
##                                                      build.tb))
##                        else:
##                            # New command registered.
##                            print >> peer, 'Programmed %r' % verbCode
##
##                return True
##
##            elif subCmd == 'deprogram':
##                # Crude.
##                # Decide verb name.
##                verbCode = command.options.verb_name or command.nextArg()
##                if not verbCode:
##                    print >> player, command.help()
##                    return True
##
##                try: commands = player.interpreter.commands
##                except AttributeError:
##                    print >> player, 'Not configured for commands!'
##                else:
##                    # First find the command function as named.
##                    action = commands.lookup(verbCode)
##                    if not action:
##                        print >> player, 'Unknown command: %s' % verbCode
##                    else:
##                        # Invoke removal of action.
##                        commands.remove(action)
##                        print >> player, 'Command removed:', action
##
##                return True



# Heartbeat Profiling.
from .wizard import parameters
from os.path import basename

# My implementation of profile and stats.
##    class Profile:
##        from time import time as timer
##        timer = staticmethod(timer)
##
##        def __init__(self):
##            self.stats = {}
##
##        def runcall(self, function, *args, **kwd):
##            start = self.timer()
##            try: return function(*args, **kwd)
##            finally:
##                self.stats.setdefault(function, []).append(self.timer() - start)
##
##    class Stats:
##        def __init__(self, other, stream = None):
##            if stream is None:
##                from sys import stdout as stream
##
##            self.stats = {}
##            self.stream = stream
##            self.add(other)
##
##        def add(self, other):
##            for (function, durations) in other.stats.iteritems():
##                self.stats.setdefault(function, []).extend(durations)
##
##        def text_stats(self):
##            yield 'NR    TOTAL       AVG         MAX         MIN         FILENAME:FUNCTION'
##            for (function, durations) in self.stats.iteritems():
##                co = function.func_code
##                nr = len(durations)
##                total = sum(durations)
##                total = float(total)
##                avg = total / nr
##                maxt = max(durations)
##                mint = min(durations)
##
##                yield '%-4d  %-10.10s  %-10.10s  %-10.10s  %-10.10s  %s:%s' % \
##                      (nr, '%4.6f' % total, '%4.6f' % avg,
##                       '%4.6f' % max(durations), '%4.6f' % min(durations),
##                       basename(co.co_filename), co.co_name)
##
##        def print_stats(self):
##            try: page = self.stream.page_string
##            except AttributeError:
##                for line in self.text_stats():
##                    print >> self.stream, line
##            else:
##                page('\r\n'.join(self.text_stats()) + '\r\n')

# Standard Profiling Implementation.
from profile import Profile
from pstats import Stats as stdStats

class Stats:
    # HACK: Wraps standard stats to make it comply with pstats.add_callers algorithm.
    def __init__(self, other, stream = None):
        if stream is None:
            from sys import stdout as stream

        self.stdStats = stdStats(self.EmptyStatsHolder(),
                                 stream = stream)

    class EmptyStatsHolder:
        # XXX pstats.Stats.load_stats does not allow empty stats.
        class SeeminglyNonEmptyStats(dict):
            def __bool__(self):
                return True

        stats = SeeminglyNonEmptyStats()
        def create_stats(self):
            pass

    class NonEmptyStatsHolder:
        def __init__(self, stats):
            self.stats = stats
        def create_stats(self):
            pass

    class Transform:
        def __init__(self, other):
            other.create_stats()
            self.stats = other.stats
        def create_stats(self):
            # Transform for add_callers algorithm.
            stats = self.stats
            self.stats = {}

            for (func, (cc, nc, tt, ct, callers)) in stats.items():
                # Transform callers to something add_callers can accept.
                for key in list(callers.keys()):
                    # Basically -- this scalar value is expected to be a singleton.
                    callers[key] = [callers[key]]

                # Fold it back into the stats.
                self.stats[func] = (cc, nc, tt, ct, callers)

    def add(self, other):
        self.stdStats.add(self.Transform(other))

    # Standard Stats Interface Wrapper
    @property
    def stats(self):
        return self.stdStats.stats

    def print_stats(self, *args, **kwd):
        return self.stdStats.print_stats(*args, **kwd)

    def print_callers(self, *args, **kwd):
        # Minor surgery: (XXX)
        os = self.stdStats.stats
        self.stdStats.stats = self.Transform(self.NonEmptyStatsHolder(self.stdStats.stats)).stats

        try: return self.stdStats.print_callers(*args, **kwd)
        finally:
            self.stdStats.stats = os

    def print_callees(self, *args, **kwd):
        return self.stdStats.print_callees(*args, **kwd)
    def strip_dirs(self, *args, **kwd):
        return self.stdStats.strip_dirs(*args, **kwd)
    def sort_stats(self, *args, **kwd):
        return self.stdStats.sort_stats(*args, **kwd)

# The profiling routine.
_stats_collection = Stats(Profile())

def doHeartbeatPulseWithProfiling(heartbeat, *args, **kwd):
    profiler = Profile()
    try: return profiler.runcall(heartbeat, *args, **kwd)
    finally: _stats_collection.add(profiler)

    ##    if _profile_stats_stream is None:
    ##        return heartbeat.pulse(*args, **kwd)
    ##
    ##    profiler = Profile()
    ##    try: return profiler.runcall(heartbeat.pulse, *args, **kwd)
    ##    finally:
    ##        # Extract stats, stream to file (in binary form).
    ##        profiler.create_stats()
    ##        import marshal
    ##        marhsal.dump(profiler.stats, _profile_stats_stream)
    ##        _profile_stats_stream.flush()

from io import StringIO as StringBuffer
class StatsTransferForShow:
    def __init__(self, stats):
        self.stats = stats.stats
    def create_stats(self):
        pass

    class StreamPager(StringBuffer):
        def page(self, peer):
            peer.page_string(self.getvalue())

    @classmethod
    def ShowStats(self, peer, order = None):
        stream = self.StreamPager()
        stats = Stats(self(_stats_collection),
                      stream = stream)

        stats.strip_dirs()
        if order in ['calls', 'cumulative', 'file', 'line', 'module',
                     'name', 'nfl', 'pcalls', 'stdname', 'time']:
            stats.sort_stats(order)

        stats.print_stats()
        stream.page(peer)

    @classmethod
    def ShowCallers(self, peer):
        stream = self.StreamPager()
        stats = Stats(self(_stats_collection),
                      stream = stream)

        stats.strip_dirs()
        stats.print_callers()
        stream.page(peer)

    @classmethod
    def ShowCallees(self, peer):
        stream = self.StreamPager()
        stats = Stats(self(_stats_collection),
                      stream = stream)

        stats.strip_dirs()
        stats.print_callees()
        stream.page(peer)

# Management.
class HeartbeatProfiler:
    Instance = None

    class NotAvailable(RuntimeError):
        pass

    @classmethod
    def IsEnabled(self, heartbeat):
        return self.Instance is not None

    @classmethod
    def Enable(self, heartbeat):
        if self.IsEnabled(heartbeat):
            raise self.NotAvailable

        HeartbeatProfiler.Instance = self(heartbeat)

    @classmethod
    def Disable(self, heartbeat):
        if not self.IsEnabled(heartbeat):
            raise self.NotAvailable

        self.Instance.disable()
        self.Instance = None

    @classmethod
    def ShowStats(self, *args, **kwd):
        StatsTransferForShow.ShowStats(*args, **kwd)

    @classmethod
    def ShowCallers(self, *args, **kwd):
        StatsTransferForShow.ShowCallers(*args, **kwd)

    @classmethod
    def ShowCallees(self, *args, **kwd):
        StatsTransferForShow.ShowCallees(*args, **kwd)

    # Manage instance.
    def __init__(self, heartbeat):
        self.heartbeat = heartbeat
        self.enable()

    # Fail on multithreading core.
    def disable(self):
        self.heartbeat.pulse = self.previous_pulse
    def enable(self):
        self.previous_pulse = self.heartbeat.pulse
        self.heartbeat.pulse = self

    def __call__(self, *args, **kwd):
        return doHeartbeatPulseWithProfiling(self.previous_pulse, *args, **kwd)

def getHeartbeatObject():
    from world import heartbeat
    return heartbeat

@parameters(1, None)
def doProfileHeartbeat(peer, command, arg = None):
    if command == 'on':
        try: HeartbeatProfiler.Enable(getHeartbeatObject())
        except HeartbeatProfiler.NotAvailable:
            print('Already profiling.', file=peer)
        else:
            print('Heartbeat profiling enabled.', file=peer)

    elif command == 'off':
        try: HeartbeatProfiler.Disable(getHeartbeatObject())
        except HeartbeatProfiler.NotAvailable:
            print('Already disabled.', file=peer)
        else:
            print('Heartbeat profiling disabled.', file=peer)

    elif command in ('show', 'stats'):
        try:
            if arg == 'callers':
                HeartbeatProfiler.ShowCallers(peer)
            elif arg == 'callees':
                HeartbeatProfiler.ShowCallees(peer)
            else:
                HeartbeatProfiler.ShowStats(peer, arg)

        except HeartbeatProfiler.NotAvailable:
            print('Heartbeat profiling not enabled.', file=peer)

    elif command in ('load', 'dump'):
        print('Not yet implemented.', file=peer)

# Startup.
from ph import getConfig
if getConfig('heartbeat-profiling') == 'on':
    try: HeartbeatProfiler.Enable(getHeartbeatObject())
    except HeartbeatProfiler.NotAvailable:
        pass



# Try Google ipaddr module.
try: import ipaddr as google_ipaddr
except ImportError:
    def _parse_ip(addr):
        return None
else:
    # Otherwise, such ip structures aren't supported.
    def _parse_ip(addr):
        try: return google_ipaddr.IP(addr)
        except ValueError:
            return None

def _parse_valid_ips(args):
    # Todo: unroll expression this into a generator function.
    return (a if ip is None else ip \
            # Provide parsed ip structure or string.
            for (a, ip) in \
                ((a, _parse_ip(a)) \
                 for a in args) if \
                 ip is not None or type(a) is str)

class IPAddressGroup(list):
    def __contains__(self, host):
        if type(host) is str:
            hostip = _parse_ip(host)
            for ip in self:
                if type(ip) is str:
                    if host == ip:
                        return True

                elif hostip in ip:
                    return True

        return False

    # Todo: provide explicit ranges?

    def __init__(self, *args):
        list.__init__(self, _parse_valid_ips(args))

    def add(self, address):
        if address not in self:
            self.append(address)



import struct

structUnpack = struct.unpack
structPack = struct.pack
structCalcsize = struct.calcsize

BYTEORDER = '@'

def structUnpack(format, buffer):
    def _():
        n = 0
        for c in format: # XXX digitized type buffers (arrays).
            i = n + struct.calcsize(BYTEORDER + c)
            yield struct.unpack(BYTEORDER + c, buffer[n:i])[0]
            n = i

    r = list(_())
    # if len(r) == 1:
    #     return r[0]

    return r

def structCalcsize(format):
    a = 0
    #import pdb; pdb.set_trace()
    for c in format:
        a += struct.calcsize(BYTEORDER + c)

    return a

class BuildRecordClass(object):
    def __new__(self, name, bases, values):
        cls = type(name, bases, values)

        field_data = values.get('Fields')
        if field_data is not None:
            cls._format = self.__build_format(field_data)
            cls._format_size = structCalcsize(cls._format)

            cls._names = self.__build_names(field_data)
            cls._defaults = self.__build_defaults(field_data)

        return cls

    @staticmethod
    def __build_format(fields):
        return ''.join(f[1].FormatString() for f in fields)

    @staticmethod
    def __build_names(fields):
        return [f[0] for f in fields]

    @classmethod
    def __build_defaults(self, field_data):
        return list(map(self.__get_field_default, field_data))

    @staticmethod
    def __get_field_default(field):
        if len(field) > 2:
            return field[2]

        return field[1].DefaultValue()

class Record(metaclass=BuildRecordClass):
    # How to pass this down to subclasses?
    @classmethod
    def FormatString(self):
        return self._format
    @classmethod
    def Size(self):
        return self._format_size
    @classmethod
    def Names(self):
        return self._names
    @classmethod
    def Defaults(self):
        return self._defaults

    # Instances.
    def __load_string(self, record):
        size = self.Size()
        if len(record) != size:
            raise ValueError('Size of record buffer (%d) != format size (%d)' % \
                             (len(record), size))

        values = structUnpack(self.FormatString(), record)
        for (i, (n, f)) in enumerate(self.Fields):
            v = values[i]
            if isinstance(f, (Record.Array, Record, Record.String)):
                # Selectively unpack -- because it's already been done by the
                # record format string, except for arrays and subrecords.
                v = f.Unpack(v)

            setattr(self, n, v)

    def __load_null(self):
        names = self.Names()
        defaults = self.Defaults()
        for x in range(len(names)):
            setattr(self, names[x], defaults[x])

    def __init__(self, record = None):
        if isinstance(record, str):
            self.__load_string(record)
        elif record is None:
            self.__load_null()
        else:
            raise TypeError('Unknown record source type (%s)' % type(record).__name__)

    @classmethod
    def Unpack(self, record):
        return self(record)

    def __iter__(self):
        return (getattr(self, name) for name in self.Names())
    def __pack__(self):
        values = tuple(self)
        return structPack(self.FormatString(), *values)

    # Types.
    @staticmethod
    def Padding(size):
        return '%dx' % size

    # XXX Err, record should subclass type..?
    class Type:
        def __init__(self, format, default = None):
            self.format = format
            self.default = default
            self.size = structCalcsize(format)

        def ArrayOf(self, size):
            return Record.Array(self, size)
        def __mul__(self, size):
            return self.ArrayOf(size)
        def __getitem__(self, size):
            return self.ArrayOf(size)

        def FormatString(self):
            return self.format
        def DefaultValue(self):
            return self.default
        def Size(self):
            return self.size

        def Unpack(self, value):
            return structUnpack(self.FormatString(), value)[0]

        def __str__(self):
            return '%s(%r)' % (self.__class__.__name__,
                               self.FormatString())
        __repr__ = __str__

    class Array(Type):
        def __init__(self, rtype, size):
            self.type = rtype
            self.size = size

            if isinstance(rtype, type):
                # Pack it into a string.
                format = '%ds' % (size * structCalcsize(rtype.FormatString()))
            elif issubclass(getattr(rtype, '__class__', None), Record.Type):
                format = '%d%s' % (size, rtype.FormatString())
            else:
                raise TypeError(type(rtype).__name__)

            Record.Type.__init__(self, format)

        def Unpack(self, value):
            return list(map(self.type.Unpack, slices(value, self.type.Size())))

        def __str__(self):
            return '%s[%s]' % (self.type, self.size)
        __repr__ = __str__

    class String(Type):
        # Was defined as an Array-subclass, but there are no requirements for that.
        def __init__(self, size):
            self.size = size
            Record.Type.__init__(self, '%ds' % size)

        def Unpack(self, value):
            return value

    class CString(String):
        def Unpack(self, value):
            return fromCString(value)

    Pointer = Type('P')

    # Todo: signs.
    Byte  = Type('b')
    Int   = Type('i')
    Short = Type('h')
    Long  = Type('l')
    UInt  = Type('I')

    SByte = Byte
    Char  = Byte
    ShInt = Short
    TimeT = Long
    SizeT = UInt

def pack(record):
    # if issubclass(record.__class__, Record):
    return record.__pack__()

def slices(array, size):
    i = 0
    end = len(array)

    while i < end:
        n = i + size
        yield array[i:n]
        i = n

def fromCString(value):
    i = value.find('\x00')
    if i < 0:
        return value

    return value[:i]

def ReadFile(fileSource, recordClass, nitems = -1):
    size = recordClass.Size()
    n = 0
    try:
        while True:
            buf = fileSource.read(size)
            if buf == '':
                break

            yield recordClass(buf)

            # Limit number of items read if specified.
            if nitems > 0:
                n += 1
                if n >= nitems:
                    break

    except EOFError:
        pass

def ReadOneFromFile(fileSource, recordClass):
    return next(ReadFile(fileSource, recordClass, 1))

def ReadBytes(fp, nr_bytes):
    return fp.read(nr_bytes)



# Player Event Tracking
from stuphos.runtime.facilities import Facility
from stuphos.runtime import newComponent
from stuphos import getConfig

# XXX dependency
from stuphos.system.api import mudlog

from xdrlib import Packer
from time import time as now

def timestamp():
    # Return time in milliseconds
    # XXX Just return 32bit time
    # return now() * 1000
    return int(now())

class BinaryLog(Packer):
    class Record(object):
        def pack(self, db):
            # Pack binary form into db.
            pass

    @classmethod
    def OpenFile(self, filename):
        return open(filename, 'w')

    # Much of the higher-level relational data will need to be kicked up into appl.
    def __init__(self, stream):
        self._Packer__buf = stream

    def flush(self):
        self._Packer__buf.flush()

    SHORT_MAX = (2 ** 16)
    INT_MAX = (2 ** 32)

    def convertFlags(self, these):
        bitv = 0
        if these is not None:
            for f in these:
                bitv |= self.FLAG_MAP[f.lower()]

        return bitv

    ##    TIMEFRAME_SIZE = 0
    ##
    ##    def convertTimestamp(self, timestamp):
    ##        return timestamp - (self.TIMEFRAME_SIZE, self._timeframe)
    ##
    ##    def activelyConvertTimestamp(Self, timestamp):
    ##        while True:
    ##            timestamp = self.convertTimestamp(record.timestamp)
    ##            timestamp *= 1000 # micros
    ##
    ##            if timestamp <= self.INT_MAX:
    ##                return timestamp
    ##
    ##            # push timeframe forward, reconvert.
    ##            self._timeframe += 1
    ##            self.pack_timeframe(self, self._timeframe)
    ##
    ##    def getPlayerIdnum(self, name):
    ##        try: return self._idnum_map[name]
    ##        except KeyError:
    ##            def newId():
    ##                s = len(self._idnum_map)
    ##                while s in self._idnum_map:
    ##                    s += 1
    ##
    ##                return s
    ##
    ##            i = self._idnum_map[name] = newId()
    ##            return i

    def buildPktHdr(self, type, flags, size):
        if isinstance(type, int):
            assert type >= 0 and type <= 255
        else:
            type = self.TYPE_MAP[type.lower()]

        if flags is None:
            flags = 0
        elif isinstance(flags, int):
            assert flags >= 0 and flags <= 255
        else:
            type = self.convertFlags(flags)

        self.pack_uint((type << 24) | (flags << 16) | size)

    def pack_packet(self, type, flags, data):
        n = len(data)
        pkthdr = self.buildPktHdr(type, flags, n)
        self.pack_fopaque(n, data)
        self.flush()

class RecordKeeping(BinaryLog):
    # Log game events, to listeners and/or to binary file.
    TYPE_EPOCH       = 10
    TYPE_TIMEFRAME   = 11
    TYPE_HOSTRECORD  = 12
    TYPE_NAMERECORD  = 13
    TYPE_PLAYERINPUT = 14
    TYPE_PEERINPUT   = 15

    BIT_LONG = (1 << 2)

    TYPE_MAP = dict(epoch         = TYPE_EPOCH,
                    time_frame    = TYPE_TIMEFRAME,
                    host_record   = TYPE_HOSTRECORD,
                    # player_record = TYPE_PLAYERRECORDS,

                    input         = TYPE_PLAYERINPUT,
                    peer_input    = TYPE_PEERINPUT)

    FLAG_MAP = dict(int = BIT_LONG)

    class Record(BinaryLog.Record):
        def present(self, peer):
            # Show this record to a player.
            pass

    def __init__(self, dbpath):
        BinaryLog.__init__(self, BinaryLog.OpenFile(dbpath))
        self.listeners = []

    # Todo: adapt log-form to message listener (pentacle)
    def addListener(self, o):
        if o not in self.listeners:
            self.listeners.append(o)
            return True

    def removeListener(self, o):
        if o in self.listeners:
            self.listeners.remove(o)
            return True

    def broadcast(self, record):
        for o in self.listeners:
            record.present(o)

    def handle(self, record):
        record.pack(self)
        self.broadcast(record)

    def __lshift__(self, record):
        self.handle(record)
        return self

class ComponentRecord(RecordKeeping.Record):
    @classmethod
    def Bind(self, component):
        for name in dir(component):
            try: obj = getattr(component, name)
            except AttributeError: pass
            else:
                try:
                    if issubclass(obj, ComponentRecord):
                        # Dynamically connect record type to instrument composing with it.
                        obj.Component = component

                except TypeError:
                    pass

        return component

    def __new__(self, ctlr, *args, **kwd):
        inst = RecordKeeping.Record.__new__(self)
        inst.__init__(*args, **kwd)

        self.Component << inst

# Eventually:
# from .behavior import Instrumentation
# from . import ComponentRecord

class Instrumentation:
    def __init__(self):
        self._hostname_map = dict()

    def newHostId(self, host):
        hostId = crc32(host)
        p = Packer()
        p.pack_int(hostId)
        p.pack_string(host)
        self.pack_packet('host_record', None, p.get_buffer())
        return hostId

    def pack_host(self, host):
        try: return self._hostname_map[host]
        except KeyError:
            hostId = self.newHostId(host)
            self._hostname_map[host] = hostId
            return hostId

    class onNewConnection(ComponentRecord):
        def __init__(self, peer):
            self.peer = peer
            self.host = peer.host

        def pack(self, db):
            db.pack_host(self.host)

        def present(self, peer):
            print('&B[&WConnect: &w%s&B]&N' % self.peer)

    class onPlayerInput(ComponentRecord):
        def __init__(self, peer, input):
            if peer.avatar is None:
                self.host = peer.host
                self.idnum = None
            else:
                self.host = None
                self.idnum = peer.avatar.idnum

            self.input = input
            self.peer = peer

        def pack(self, db):
            if self.idnum is None:
                # This must be done before any new record.
                hostId = db.pack_host(self.host)

                p = Packer()
                p.pack_uint(hostId)
                p.pack_uint(timestamp())
                p.pack_string(self.input)
                db.pack_packet('peer_input', None, p.get_buffer())
            else:
                p = Packer()
                p.pack_uint(timestamp())
                p.pack_uint(self.idnum)
                p.pack_string(self.input)
                db.pack_packet('input', None, p.get_buffer())

        def present(self, peer):
            print('&B[&w%s&B] &N%s&N' % (self.peer, self.input), file=peer)

    ##    class onMovement(ComponentRecord):
    ##        def __init__(self, mobile, origin, destination, direction, move_type, move_cost):
    ##            self.mobile = mobile
    ##
    ##        def pack(self, db):
    ##            if not self.mobile.npc:
    ##                db.pack()

    ##    class onSlaying(ComponentRecord):
    ##        def __init__(self, victim, killer):
    ##            # And also, for instance, experience earned (approx), level at the time
    ##            pass

class PlayerTracking(Facility, Instrumentation, RecordKeeping):
    '''
    [Facilities]
    facility.tracking: mud.management.tracking.PlayerTracking

    [PlayerTracking]
    db-path: etc/player-tracking.db
    '''

    NAME = 'Player::Tracking'
    AUTOMANAGE = True

    class Manager(Facility.Manager):
        MINIMUM_LEVEL = Facility.Manager.IMPLEMENTOR
        VERB_NAME = 'player-tracking*'

        def do_listen(self, peer, cmd, args):
            t = PlayerTracking.get()
            if t is not None:
                if t.addListener(peer):
                    print('You will now be informed of player events.', file=peer)
                else:
                    print('You are already listening to player events.', file=peer)

        def do_unlisten(self, peer, cmd, args):
            t = PlayerTracking.get()
            if t is not None:
                if t.removeListener(peer):
                    print('You will no longer be informed of player events.', file=peer)
                else:
                    print('You are not currently listening to player events.', file=peer)

        def do_last(self, peer, cmd, args):
            # Show all records from log since time.
            pass

    @classmethod
    def create(self):
        dbpath = getConfig('db-path', section = 'PlayerTracking')
        if not dbpath:
            raise ValueError('PlayerTracking:db-path is not configured')

        return ComponentRecord.Bind(newComponent(self, dbpath = dbpath))

    def __instance_init__(self):
        Instrumentation.__init__(self)
        RecordKeeping.__init__(self, self.dbpath)

    # Er, do this onShutdownGame instead??
    ##    def __registry_delete__(self):
    ##        rotated = '%s.%s' % (self.dbpath, now())
    ##        rename(self.dbpath, rotated)
    ##        system('gzip -9 %r' % rotated)


PlayerTracking.get(create = True)
PlayerTracking.manage()

class NSA(Instrumentation.onPlayerInput):
    # Called as a special procedure.
    def __new__(self, player, *args, **kwd):
        if not player.npc:
            t = PlayerTracking.get()
            if t is not None:
                inst = RecordKeeping.Record.__new__(self)
                inst.__init__(player, *args, **kwd)

                t << inst

    def __init__(self, player, this, cmd, argstr):
        self.idnum = player.idnum
        self.host = '' if player.peer is None else player.peer.host
        self.input = ' '.join([_f for _f in (cmd.name, argstr) if _f])
        self.peer = player.peer



# Database connection configuration service.
#
from contextlib import contextmanager
import json

class ConfigurationError(ValueError):
	pass

class DBCore(dict):
	'''
	[DBCore]
	primary.type = pg-auth
	primary.path = db.conf
	primary.host = 127.0.0.1
	primary.port = 5432
	secondary.type = sqlite
	secondary.path = sqlite:/etc/db.sqlite

	'''

	class PGAuthConfiguration:
		ENGINE = 'django.db.backends.postgresql_psycopg2'

		DEFAULT_PG_HOST = 'localhost' # '127.0.0.1'
		DEFAULT_PG_PORT = 5432

		def __init__(self, **kwd):
			self.path = kwd['path'] # assert exists
			self.host = kwd.get('host') #, self.DEFAULT_PG_HOST)
			self.port = kwd.get('port') #, self.DEFAULT_PG_PORT)

		@property
		def connectionString(self):
			if self.host or isinstance(self.port, int) or \
			    isinstance(self.port, str) and self.port.isdigit():
				return buildPGINetConnectionURIAuto \
                                    (authfile = self.path,
                                     port = self.port,
                                     host = self.host)

			return buildPGLocalConnectionURIAuto(authfile = self.path)

		@property
		def djangoDatabaseConfig(self):
			(dbName, username, password) = readPGAuthAuto(self.path)

			cfg = {'NAME': dbName,
			       'ENGINE': self.ENGINE,
			       'USER': username,
			       'PASSWORD': password}

			if self.host or isinstance(self.port, int) or \
			   isinstance(self.port, str) and self.port.isdigit():
			    cfg['HOST'] = self.host or self.DEFAULT_PG_HOST
			    cfg['PORT'] = int(self.port)

			return cfg

		@property
		def fields(self):
			return dict(path = self.path,
						host = self.host,
						port = self.port)

	class SqliteConfiguration:
		ENGINE = 'django.db.backends.sqlite3'

		def __init__(self, **fields):
			self.filepath = fields.get('file-path')

			try: self.path = fields['path']
			except KeyError:
				if self.filepath is None:
					raise ConfigurationError('Sqlite databases must define at least file-path')

				self.path = f'sqlite:{self.filepath}'

		@property
		def connectionString(self):
			return self.path

		@property
		def djangoDatabaseConfig(self):
			return dict(NAME = self.filepath,
				        ENGINE = self.ENGINE,
				        USER = '',
				        PASSWORD = '')

		@property
		def fields(self):
			return dict(path = self.path,
						filepath = self.filepath)

	ConfigurationTypes = {'pg-auth': PGAuthConfiguration,
						  'pg': PGAuthConfiguration,
						  'postgres': PGAuthConfiguration,
						  'sqlite': SqliteConfiguration}

	def buildConfiguration(self, **fields):
		config = self.ConfigurationTypes[fields['type']]
		return config(**fields)

	def installConfiguration(self, name, **fields):
		cfg = self[name] = self.buildConfiguration(**fields)
		return cfg

	@classmethod
	def Load(self, section = 'DBCore'):
		# This is here for running web migrate --resync-db
		#import pdb; pdb.set_trace()

		from stuphos import getSection
		cfg = getSection(section)
		assert cfg is not None

		dbs = dict()

		for o in cfg.options():
			(name, field) = o.split('.', 1)
			d = dbs.setdefault(name, dict())
			d[field] = cfg.get(o)

		return self.LoadFromDBConfig(dbs)

	@classmethod
	def LoadFromDBConfig(self, dbs):
		# Validate while building.
		core = self()
		for (name, fields) in dbs.items():
			core.installConfiguration(name, **fields)

		return core

	def InitializeForThread(self, name = None):
	    self.hub.threadConnection = None if name is None else self.getConnection(name)
	def InitializeForProcess(self, name = None):
	    self.hub.processConnection = None if name is None else self.getConnection(name)

	@property
	def hub(self):
	    from sqlobject import sqlhub
	    return sqlhub


	class ConnectionDescriptor:
		def __init__(self, connectString):
			self.connectString = connectString

		def __call__(self):
			return self.connectString


	def getConnection(self, name):
		if isinstance(name, self.ConnectionDescriptor):
			connect = name()
		else:
			connect = self[name].connectionString

		from sqlobject import connectionForURI
		return connectionForURI(connect)

	def getDjangoDatabaseConfig(self, name):
	   	return self[name].djangoDatabaseConfig

	@contextmanager
	def hubThread(self, name):
		hub = self.hub
		try: prev = hub.threadConnection
		except AttributeError: prev = None

		conn = hub.threadConnection = self.getConnection(name)

		try: yield conn
		finally:
			hub.threadConnection = prev

	@contextmanager
	def hubProcess(self, name):
		hub = self.hub
		try: prev = hub.processConnection
		except AttributeError: prev = None

		conn = hub.processConnection = self.getConnection(name)

		try: yield conn
		finally:
			hub.processConnection = prev

class DynamicDBCore(DBCore):
	@classmethod
	def Load(self, master, masterNamespace = 'primary'):
		return self(master, masterNamespace)

	def __init__(self, master, masterNamespace):
		self.master = master
		self.masterNamespace = masterNamespace

	def loadConfig(self, name):
		from .orm import DBConf

		with master.hubThread(self.masterNamespace):
			for cfg in DBConf.selectBy(name = name):
				return self.buildConfiguration(**json.loads(cfg.fields))

	def saveConfig(self, name, cfg):
		from .orm import DBConf

		with master.hubThread(self.masterNamespace):
			DBConf(name = name, fields = json.dumps(cfg.fields)).save()


# PG
def readPGAuthAuto(authfile):
    # Get the configuration from file.
    with open(authfile) as dbConf:
        dbName = dbConf.readline().strip()
        username = dbConf.readline().strip()
        password = dbConf.readline().strip()

    return (dbName, username, password)

def buildPGLocalConnectionURIAuto(authfile):
    (dbName, username, password) = readPGAuthAuto(authfile)

    return 'postgres://{username}:{password}@/{database}'.format \
           (username = username, password = password, database = dbName)

def buildPGINetConnectionURIAuto(authfile, port = None, host = None):
    (dbName, username, password) = readPGAuthAuto(authfile)

    if port is None:
        port = DBCore.PGAuthConfiguration.DEFAULT_PG_PORT
    if host is None:
        host = DBCore.PGAuthConfiguration.DEFAULT_PG_HOST

    return 'postgres://{user}:{password}@{host}:{port}'.format \
           (user = username, password = password,
            host = host, port = port)

# XXX Get rid of this, if possible.  It should exist only in the runtime bootstrap.
from stuphos.runtime.registry import getRegistry
getRegistry(create = True)

# Instance.
dbCore = DBCore.Load()

def initDjangoDatabaseConfig(name):
    # A bootstrapped method for obtaining database config.
    # from stuphos.runtime.registry import getRegistry
    # getRegistry(create = True)

    return dbCore.getDjangoDatabaseConfig(name)

# dynamic_dbCore = DynamicDBCore.Load(dbCore)



# Connects Girl interpreter to player shell/command experience.
# import new

from stuphos.kernel import Girl, Script, Volume, Undefined, Programmer
from stuphos.kernel import GirlSystemModule, Machine as VM, GrammaticalError
from stuphos.kernel import findUserByName, OuterFrame, TaskCreation, constrainStructureMemory
from stuphos.kernel import resolve_procedure

from stuphos.runtime import Object
from stuphos import getConfig

from stuphmud.server.player import getPlayerScope, events, HandleCommandError
from stuphmud.server.player.interfaces.code import ShellFactoryBase
from stuphmud.server.adapters import PeerAdapter, MobileAdapter, TriggerAdapter

from stuphos.etc.tools import getKeyword, Option, isYesValue
from . import ProgrammeManager, EvaluationShell, Programme, isFromSecureHost

# Permission level for non-authenticated command users.
DEFAULT_PROGRAMMER = Programmer.NONE


COMMAND_TASK = "call('immersion/interpreter/command', me, command, subcommand, arguments)"
# COMMAND_TASK = "'immersion/interpreter/command'(me, command, subcommand, arguments)"

VERB_TASK = "return call('immersion/interpreter/verb', me, command, arguments)"
# VERB_TASK = "return 'immersion/interpreter/verb'(me, command, arguments)"

WEB_COMMAND_TASK = "return 'immersion/interpreter/interpretAgentWeb'" + \
    '(security$context$new().resident, locals$(), session, script, syntax, parameters)'

LOCATION_UPDATE_TASK = "return 'immersion/interpreter/locationUpdate'" + \
    '(security$context$new(), locals$(), player, locationUpdate)'


class EvaluateCodeTaskCreation(TaskCreation):
    def __init__(self, peer = None, completion = None, initialize = None,
                 *args, **kwd):
        TaskCreation.__init__(self, *args, **kwd)

        self.peer = peer
        self.completion = completion
        self.initialize = initialize
        # self.environ = environ

        # debugOn()
        if peer is None:
            self.programmer = DEFAULT_PROGRAMMER
            self.user = None
        else:
            name = peer.avatar.name
            self.programmer = Programmer(name)
            self.user = findUserByName(name)

        self.taskClassKwd = dict(user = self.user)

    def createTask(self, environ):
        task = TaskCreation.createTask(self, environ)
        task.environ.update(**self.environScope)

        # print(f'[evalCode.createTask] {dict.keys(task.environ)}')

        self.procedure.setEnvironment(task.environ)

        return task

    def postInitTask(self, task, frame):
        user = self.user
        if user is not None:
            self.locals._connectMemory(VM.Task.Memory.Connect(task, user))

        if self.completion:
            frame.onComplete(self.completion)

        if self.initialize:
            with self.vm.threadContext(task = task):
                self.initialize(task, frame)

        return TaskCreation.postInitTask(self, task, frame)


# Implementation of VM.Task.
class PlayerScript(Script):
    class BadPeerState(ValueError):
        pass

    def __init__(self, peer, shell, tracing, *args, **kwd):
        self.peer = peer
        self.shell = shell

        Script.__init__(self, *args, **kwd)

        self.uncaughtError = self.handleError

        if tracing:
            # debugOn()
            if isYesValue(configuration.Interpreter.console_operator) or tracing == 'console':
                self.tracing = self.traceToConsole
            else:
                self.tracing = self.debugInstructionPeer(self.peer)

            if tracing == 'full-stack' and isYesValue(configuration.AgentSystem.allow_full_stack):
                self.debugStack = 'full'
            if tracing in ['stack', 'console']:
                self.debugStack = True


    def frameOneForPeer(self, peer, *args, **kwd):
        return Script.frameOne(self, *args, **kwd)

    def frameOne0(self, *args, **kwd):
        # if 0:
        #     # :console: :debugging:
        #     import sys
        #     print >> sys.stderr, 'Frames:', len(self.frames)
        #     # print >> sys.stderr, '\n'.join(str(f.procedure) for f in self.frames)
        #     p = self.frame.procedure
        #     if hasattr(p, 'position'):
        #         o = p.position(0)
        #         print >> sys.stderr, p.instructionsString(o, o+1)

        # Swap in peer for console.
        # try: return self.shell.withPeerHeadAndException(self.peer, (self.Done, VM.Yield),
        #                                                 self.frameOneForPeer,
        #                                                 *args, **kwd)
        # # except Script.Yield:
        # #     import sys
        # #     (etype, value, tb) = sys.exc_info()
        # #     raise (etype, value, tb)

        # except Script.Done, e:
        #     # This is not necessary in all cases of Script.Done at this level.
        #     # This is because Done is also raised on exception, in which case
        #     # the stack is not REPL-integrated.
        #     # try: result = self.stack.pop()[0]
        #     # except IndexError: pass
        #     # else:
        #     #     if result is not None:
        #     #         print >> self.peer, repr(result)

        #     raise e

        return self.shell.withPeerHeadAndException \
                    (self.peer, (self.Done, VM.Yield),
                     self.frameOneForPeer,
                     *args, **kwd)

    def handleError(self, task, frame, exc, traceback):
        # XXX Emitting no error for some things but still raising Done.
        (etype, value, tb) = exc
        if isinstance(value, VM.Yield):
            return # This may not be right since it might need to be propogated.
            raise etype(value).with_traceback(tb)

        if isYesValue(getConfig('native-traceback', 'Interpreter')):
            HandleCommandError(self.peer, exc, frame_skip = 0)

        self.peer.sendTraceback(self, exc, traceback)

        # Eventually this will go away in favor of traceback logging.
        if isYesValue(getConfig('report-player-error', 'Interpreter')):
            from stuphos import logException
            logException(etype = etype, value = value, tb = tb, traceback = True,
                         header = f'[player$script$error] {etype.__name__}: {value}')

        self.logTraceback(task, traceback)
        raise self.Done


    traceToConsole = Script.debugInstruction

    # def traceToConsole(self, frame, pos, instr, args):
    #     # debugOn()
    #     task = frame.task

    #     try: name = frame.procedure.getSourceMap().source
    #     except AttributeError:
    #         name = task.name

    #     instr = getattr(instr, '__name__', '?')
    #     args = ', '.join(map(str, args))

    #     msg = '%-20.20s %04d %s(%s)' % (name, pos, instr, args)

    #     print(msg)
    #     if task.stack:
    #         # Don't print anything if there's no stack, otherwise
    #         # it will look like an empty string on the stack.
    #         print('    ' + '\n    '.join(map(str, task.stack)))

    def handleVerbOutcome(self, outcome, command, argstr):
        # Override this in a new script class for custom command recognition.
        if not outcome:
            print('Unknown command: %r' % command, file=self.peer)


    # Invocation Methods:
    @classmethod
    def evaluateCodeAsync(self, peer, shell, program, scope, completion,
                          tracing = False, check_active = True, initialize = None,
                          taskCreation = EvaluateCodeTaskCreation,
                          **environ):

        # Because this code starts and runs code with a certain authorization, we make
        # a determination in our security model if that authority passes the check.

        if peer.state not in ['Playing', 'Shell']:
            # Must be logged in or authenticated.
            raise self.BadPeerState

        if isFromSecureHost(peer):
            kwd = dict(peer = peer, vm = getVirtualMachine(), procedure = program,
                       taskClass = self, taskClassArgs = (peer, shell, tracing),
                       locals = scope, completion = completion, environ = environ,
                       initialize = initialize, checkActiveTasks = check_active)

            return taskCreation.Create(**kwd)

        else:
            logOperation(f'[player$evaluate] Bad peer source: {peer}')


    # @classmethod
    # def evaluateCodeAsync(self, peer, shell, program, scope, completion,
    #                       tracing = False, check_active = True, **environ):

    #     # Because this code starts and runs code with a certain authorization, we make
    #     # a determination in our security model if that authority passes the check.
    #     if isFromSecureHost(peer):
    #         # from stuphos.etc import isYesValue
    #         # from stuphos import getConfig

    #         if peer.avatar is None:
    #             progr = DEFAULT_PROGRAMMER
    #             user = None
    #         else:
    #             name = peer.avatar.name
    #             progr = Programmer(name)
    #             user = findUserByName(name)


    #         # Start new PlayerScript task.
    #         if check_active:
    #             # This might have been done before compiling the program.
    #             from stuphos.kernel import checkActiveTasks
    #             checkActiveTasks(user)

    #         task = self(peer, shell, tracing, user = user)
    #         task.environ.update(**environ)

    #         # This is for a persistance concept that doesn't yet exist so it should be removed.
    #         # task.environ['book'] = Volume(getattr(peer, 'environ', task.environ), program) # or scope

    #         # if isYesValue(getConfig('system-module', 'AgentSystem')):
    #         #     task.environ['system'] = GirlSystemModule.Get()

    #         program.setEnvironment(task.environ)


    #         vm = getVirtualMachine()
    #         new = task.addFrameCall(program, programmer = progr)
    #         new.locals = scope
    #         if user is not None:
    #             # Note: scope must be a managed memory object.
    #             # Todo: make it vm.Task...?
    #             scope._connectMemory(VM.Task.Memory.Connect(task, user))

    #         if completion:
    #             # print(f'registering completion: {completion}')
    #             new.onComplete(completion)

    #         vm += task
    #         return task

    class reprCompletion:
        def __init__(self, peer):
            self.peer = peer

        def __call__(self, frame, *error):
            task = frame.task
            if task.stack and error[1] is None:
                try: value = task.stack.pop()[0]
                except IndexError: pass
                else: self.displayValue(value)

        def displayValueToPeer(self, reprValue):
            if isinstance(reprValue, str):
                # todo: move actual printing to method on peer.
                # print(f'task return value: {value}')
                print(reprValue, file = self.peer)

        def displayValue(self, value):
            if value is not None:
                # debugOn()

                # XXX This is unsafe when arbitrarily importing python objects (that may define __getattr__).
                try: _reprc = value._reprOrContinuation
                except AttributeError as e:
                    # print(f'display value _reprc: {e}')
                    value = repr(value)
                else:
                    # If this returns None, there was a continuation.
                    value = self.runDisplayValue(_reprc, value)

                self.displayValueToPeer(value)

        def runDisplayValue(self, _reprc, value):
            # return repr(value)

            # Run the repr method in this task!

            if callable(_reprc):
                try: return _reprc()
                except OuterFrame as sub:
                    @sub.onComplete
                    def reprCompleted(frame, *error):
                        if error[1] is None:
                            self.displayValueToPeer(frame.task.stack.pop()[0])

                # except Exception as e:
                #     print(f'[display] {e.__class__.__name__}: {e}')
                #     raise e


        # elif isinstance(error[1], task.frames.ForeignFrameError):
        #     frame = error[1].frame
        #     print >> peer, 'foreign:', frame #.procedure

    @classmethod
    def evaluateCode(self, peer, shell, program, scope, tracing = False, **environ):
        return self.evaluateCodeAsync(peer, shell, program, scope,
                                      self.reprCompletion(peer),
                                      tracing = tracing, **environ)

    @classmethod
    def evaluateInput(self, peer, type, input, tracing = False, shell = None):
        return self.evaluateInputAsync(peer, type, input, self.reprCompletion(peer),
                                       tracing = tracing, shell = shell)

    @classmethod
    def evaluateInputAsync(self, peer, type, input, completion, tracing = False, shell = None):
        scope = getGirlPlayerScope(peer) # target of @dir()
        program = Girl(getattr(Girl, type.capitalize()), input)

        if shell is None:
            shell = peer.interpreter

        if isYesValue(configuration.Interpreter.dump_expression_code):
            peer.sendln(program.instructionsString())

        scope.update(this = PeerAdapter(peer), # todo: persist these
                     me = MobileAdapter(peer.avatar))

        self.evaluateCodeAsync(peer, shell, program, scope, completion,
                               tracing = tracing)

        return True

    @classmethod
    def evaluateStatement(self, peer, statement, tracing = False, shell = None):
        return self.evaluateInput(peer, 'statement', statement, tracing = tracing, shell = shell)

    @classmethod
    def evaluateScript(self, peer, script, tracing = False, shell = None):
        return self.evaluateInput(peer, 'module', script, tracing = tracing, shell = shell)


    # Command/Verb invocation: the interpreter is a running task that is stored
    # as a weakref on the peer object, started by this class if it is not already
    # running (evaluates as non-None).  Commands and verbs are merely objects
    # passed to this running task using a synchronization queue.  The interpreter
    # program is responsible for sequencing the synchronicity of each command,
    # as well as (re)drawing the prompt.

    @classmethod
    def evaluateCommand(self, actor, name, subcmd, argstr):
        peer = actor.peer
        if peer is not None:
            scope = getGirlPlayerScope(peer) # target of @dir()
            code = configuration.Interpreter.command_task or COMMAND_TASK

            try: program = Girl(Girl.Module, code + '\n')
            except GrammaticalError as e:
                print(f'{code}:')
                print(e.report())
            else:
                scope.update(subcommand = subcmd, command = name,
                             arguments = argstr,
                             this = PeerAdapter(peer),
                             me = MobileAdapter(actor))

                self.evaluateCode(peer, peer.interpreter, program, scope,
                                  tracing = getGirlTracingLevel(peer))

    @classmethod
    def evaluateVerb(self, peer, actor, command, argstr):
        scope = getGirlPlayerScope(peer) # target of @dir()
        code = configuration.Interpreter.verb_task or VERB_TASK

        try: program = Girl(Girl.Module, code + '\n')
        except GrammaticalError as e:
            print(f'{code}:')
            print(e.report())
        else:
            # A command that has entered evaluate-verb mode has abandoned all
            # hope of getting output any time soon (before the end of this
            # heartbeat), because evaluateCode initiates a new asynchronous task.
            peer.hasPrompt = True # If a command doesn't generate any output, it must set this false.

            scope.update(command = command,
                         arguments = argstr,
                         this = PeerAdapter(peer),
                         me = MobileAdapter(actor))

            task = self.evaluateCode(peer, peer.interpreter, program, scope,
                                     tracing = getGirlTracingLevel(peer))

            if task is not None:
                # XXX Since using evaluateCode means automatically using a
                # reprCompletion, this completion will fail (silently).
                @task.onComplete
                def completion(_, exception = None, traceback = None):
                    if exception is None:
                        try: outcome = task.stack.pop()[0]
                        except IndexError: pass # Shouldn't happen (see above)
                        else: task.handleVerbOutcome(outcome, command, argstr)

                # Consider this command handled.
                return True

    @classmethod
    def evaluateWebCommand(self, peer, interpreter, scope, completion,
                           sessionAdapter, script, syntax, params,
                           tracing = False, **environ):

        # XXX completion not used.

        code = configuration.Interpreter.web_command_task or WEB_COMMAND_TASK

        try: program = Girl(Girl.Module, code + '\n')
        except GrammaticalError as e:
            print(f'{code}:')
            print(e.report())
        else:
            # debugOn()
            if not isinstance(params, dict):
                params = dict()

            # Make a copy for interpreter logging.
            paramsIncoming = params.copy()

            params.update(environ)
            params['paramsIncoming'] = paramsIncoming

            scope.update(session = sessionAdapter,
                         script = script,
                         syntax = syntax)

            scope['instance$'] = sessionAdapter

            def paramsInit(task, frame):
                scope['parameters'] = task.Environment(task.memory, params)

            return self.evaluateCode(peer, peer.interpreter, program, scope,
                                     tracing = getGirlTracingLevel(peer),
                                     initialize = paramsInit)


    @classmethod
    def evaluateEvent(self, peer, code, tracing = False, **environ):
        scope = getGirlPlayerScope(peer)

        try: program = Girl(Girl.Module, code + '\n')
        except GrammaticalError as e:
            print(f'{code}:')
            print(e.report())
        else:
            task = self.evaluateCode(peer, peer.interpreter, program, scope,
                                     tracing = tracing, **environ)

            if task is not None:
                @task.onComplete
                def eventComplete(_, exception = None, traceback = None):
                    # debugOn()
                    if exception is not None and exception[1] is not None:
                        # Uncaught error.
                        peer.sendTraceback(task, exception, traceback)
                        # return True

                return True


    @classmethod
    def evaluateMethodCall(self, request, path, name, *args):
        # RPC.  See phsite.network.adapter.commands.doCallMethod alternatively.
        # todo: finish:
        #   instantiate self for task class
        #   interface with isFromSecureHost and request/peer
        #   integrate with SessionManager
        #   this came from mental.library.model.GirlCore.rpcCallMethod
        from stuphos.kernel import checkActiveTasks, Programmer, nullproc, getLibraryCore, protectedMemoryLoad

        # name = path[-1]
        # path = path[:-1]

        # Acquire programmer and task state.
        progr = Programmer.NONE
        if request.user is not None:
            checkActiveTasks(request.user)

            for player in request.user.default_players.all():
                progr = Programmer(player.player.player_name)
                break

        task = Script.Load(user = request.user)
        core = getLibraryCore(task)

        vm = runtime[runtime.System.Engine]
        args = protectedMemoryLoad(task, args)

        task += dict(procedure = nullproc(), programmer = progr)
        vm += task

        return callGirlMethod(core, task, path, name, *args) () # block.


    @classmethod
    def evaluateLocationUpdate(self, session, ch):
        try: locationUpdate = agentSystem[f'assets/{session.peer.avatar.name}/session$ui'] \
                .structureCached.locationUpdate

        except (NameError, KeyError, AttributeError):
            session.postMessages(['location-update', ch])
            return

        # assert isinstance(program, Procedure)
        # program = resolve_procedure(program)


        code = configuration.Interpreter.location_update_task or LOCATION_UPDATE_TASK

        try: program = Girl(Girl.Module, code + '\n')
        except GrammaticalError as e:
            print(f'{code}:')
            print(e.report())
            return


        # from ph.emulation.operation.kernel import ProgramBuffer
        # from ph.emulation.operation.instructions import Code
        # from ph.interpreter.mental.kernel import Unit

        # class ReturnLocationInfo(ProgramBuffer, Code, Unit):
        #     pass

        # program = ReturnLocationInfo()
        # program.extend([[program.push_literal, [ch]],
        #                 [program.return_from, []]])


        class UpdateLocationTaskCreation(EvaluateCodeTaskCreation):
            def preInitTask(self, task, *args, **kwd):
                nonlocal ch # The reason this isn't in module scope.
                task.environ['ch'] = constrainStructureMemory(task, ch)
                return EvaluateCodeTaskCreation.preInitTask(self, task, *args, **kwd)

        def updateCompletion(frame, etype = None, value = None, tb = None, *error):
            nonlocal ch # The reason this isn't a classmethod.
            if value is None:
                import json
                ch = frame.task.stack.pop()[0]
                ch = json.loads(json.dumps(ch))

            session.postMessages(['location-update', ch])


        peer = session.peer
        scope = getGirlPlayerScope(peer)

        scope.update(player = MobileAdapter(peer.avatar),
                     locationUpdate = locationUpdate)

        # debugOn()
        task = self.evaluateCodeAsync(peer, peer.interpreter, program, scope,
                                      updateCompletion,
                                      taskCreation = UpdateLocationTaskCreation,
                                      tracing = getGirlTracingLevel(peer))

        if task is not None:
            task.name = 'location-update'

        return task


    @classmethod
    def invokeTrigger(self, peer, program, tracing = False):
        # This only makes sense for a player.
        scope = getGirlPlayerScope(peer)
        scope.update(player = MobileAdapter(peer.avatar),
                     trigger = TriggerAdapter(self))

        self.evaluateCode(peer, peer.interpreter, program, scope,
                          tracing = tracing)

AgentScript = PlayerScript

class GirlPlayerProgrammeTrigger(Programme, events.Trigger):
    # Dual function: as stored programme, and also as triggerable player event.
    class _Meta(Programme._Meta):
        Attributes = Programme._Meta.Attributes + ['tracing']

    def __init__(self, *args, **kwd):
        tracing = getKeyword(kwd, 'tracing')
        Programme.__init__(self, *args, **kwd)
        self.tracing = tracing

    def __getstate__(self):
        state = Programme.__getstate__(self)
        state['tracing'] = self.tracing
        return state

    def getCompiledCode(self):
        if isinstance(self.sourceCode, str):
            # Compile as module.
            return Girl(Girl.Module, self.sourceCode) # self.source.replace('\r', '')

    getTriggerCode = getCompiledCode

    def invokeProgramme(self, shell, peer):
        code = self.getCompiledCode()
        scope = getGirlPlayerScope(peer)
        PlayerScript.evaluateCode(peer, shell, code, scope,
                                  tracing = self.tracing,
                                  player = peer.avatar,
                                  this = peer, me = peer.avatar)

    def invokeTrigger(self, player):
        PlayerScript.invokeTrigger(player.peer, self.getTriggerCode(),
                                   self.tracing)

class GirlCodeManager(ProgrammeManager):
    # ProgrammeManager:
    ProgrammeClass = GirlPlayerProgrammeTrigger

    def getManagerName(self):
        return 'Girl'
    def getManagerId(self):
        return '%s.%s.Instance' % (self.__class__.__module__, self.__class__.__name__)

class GirlCodeShell(EvaluationShell):
    # EvaluationShell:
    def executeGirl(self, shell, peer, argstr):
        return self.executeSourceCode(shell, peer, argstr)

    __call__ = executeGirl

    def executeCode(self, shell, peer, program):
        scope = self.getScope(peer)
        PlayerScript.evaluateCode(peer, shell, program, scope,
                                  tracing = self.tracing,
                                  player = peer.avatar,
                                  this = peer, me = peer.avatar)

    def compileSourceCodeBlock(self, peer, sourceCode):
        if isinstance(sourceCode, str):
            # Currently always is -- Compile as module.
            return Girl(Girl.Module, self.manager.formatSourceCode(sourceCode))

    def compileSingleStatement(self, peer, sourceCode):
        return Girl(Girl.Statement, self.manager.formatSourceCode(sourceCode))

    # Implementation.
    def __init__(self, manager, tracing = False):
        EvaluationShell.__init__(self, manager)
        self.tracing = tracing

    def getScope(self, peer):
        return getGirlPlayerScope(peer)

# Infrastructure.
def getVirtualMachine():
    from world import heartbeat as vm
    return vm

def getGirlScope(scope, name = None):
    # Basically return a persistant, shared namespace associated with,
    # wrapping, and being wrapped by the standard player scope.

    try: return scope.namespace
    except AttributeError: pass

    try: memory = scope.memory
    except AttributeError:
        memory = scope.memory = VM.Task.Memory.Connect(None, user = findUserByName(name))

    ns = scope.namespace = VM.Task.Environment(memory)
    return ns

    # girl = getattr(scope, 'girl', Undefined)
    # if girl is Undefined:
    #     girl = new.module('%s.namespace' % Girl.__module__)

    #     # XXX :skip: UNSAFE.
    #     girl.shared = scope
    #     scope.girl = girl

    #     # Presumably, the girl-core facility is loaded before this is initialized.
    #     # XXX :skip: UNSAFE, because it exposes the core object to all player scripts.
    #     girl.core = runtime[runtime.Girl.System]

    # return girl.__dict__

# Singleton.
GirlCodeManager.Instance = GirlCodeManager()

def getGirlPlayerScope(peer):
    name = peer.avatar.name if peer.avatar else None
    return getGirlScope(getPlayerScope(peer), name = name)
def getGirlCodeEvaluator(*args, **kwd):
    return GirlCodeShell(GirlCodeManager.Instance, *args, **kwd)

class GirlCodeEvaluatorFactory(ShellFactoryBase):
    OPTIONS = [Option('-t', '--tracing', action = 'store_true')]

    def __new__(self, peer, tracing = False):
        print(EvaluationShell.PROGRAMMING_HEADER % 'Girl', file=peer)
        return getGirlCodeEvaluator(tracing)

def getGirlTracingLevel(peer):
    try: return peer.tracing
    except AttributeError:
        tracing = configuration.Interpreter.trace_commands
        if tracing == 'stack':
            return tracing

        return isYesValue(tracing)

def doGirlStatement(peer, cmd, argstr):
    # Execute single statement now.
    # todo: security policy
    if peer.avatar and peer.avatar.implementor:
        if argstr:
            # todo: delete_doubledollar(argstr)
            try:
                PlayerScript.evaluateStatement(peer, argstr, shell = peer.interpreter,
                                               tracing = getGirlTracingLevel(peer))
            except GrammaticalError as e:
                peer.sendln(e.report())

            return True

# Command Line.
def initCommands():
    try: from stuphmud.server.player import getSharedCommands # ...plain ACMD still being initialized.
    except ImportError: pass
    else: getSharedCommands().insertOverridingAll('@', doGirlStatement)

def createCommand(verb, path, commandName):
    def performGirlCommand(peer, cmd, argstr):
        from world import heartbeat as vm
        from stuphos.kernel import Script

        core = runtime[runtime.Agent.System]
        path = path.split('/')
        node = core.root.lookup(*path)
        task = Script() # Todo: PlayerScript import

        # Open subroutine and schedule.
        try: pos = node.module.symbols[commandName][0]
        except KeyError:
            raise NameError(commandName)

        # Todo: use string-name overload
        method = node.module.getSubroutine(pos)

        # todo: use executeGirl for synchronous option.
        task.addFrameCall(method, arguments = [cmd, argstr])
        vm += task
        return True

    from stuphmud.server.player import ACMD
    return ACMD(verb)(performGirlCommand)



# Also found in etc.tools.__init__
def apply(function, *args, **kwd):
	return function(*args, **kwd)


def importAllMembers(o):
	try: return o.__all__
	except AttributeError:
		return (n for n in dir(o) if not n.startswith('_'))

def importAll(module, scope):
	from .lookup import LookupObject
	module = LookupObject(module)
	for n in importAllMembers(module):
		scope[n] = getattr(module, n)

def importAllInit(name, msg, scope):
	if name:
		try: importAll(name, scope)
		except ImportError as e:
			print(f'{msg} {e}')



"Convert rent file into Python."
from struct import calcsize, unpack

__all__=['scan_rentfile', 'get_rent']

MAX_OBJ_AFFECT=6
structfmt_obj_affected_type='2b'
structfmt_obj_file_elem    ='I128s256s128sh10il' + \
	(structfmt_obj_affected_type * MAX_OBJ_AFFECT)
structfmt_rent_info        ='14i'

structsz_obj_affected_type =calcsize(structfmt_obj_affected_type)
structsz_obj_file_elem     =calcsize(structfmt_obj_file_elem)
structsz_rent_info         =calcsize(structfmt_rent_info)

def checksize(size):
	size-=structsz_rent_info
	assert (size>=0) and not (size%structsz_obj_file_elem), \
		IOError('File not the right size!')

def scan_rentfile(path):
	# Generator
	rentfile=open(path).read() # XXX Read from file instead of slicing
	size=len(rentfile)

	checksize(size)

	yield unpack(structfmt_rent_info, rentfile[:structsz_rent_info])
	size-=structsz_rent_info

	while size>0: # >= structsz_obj_file_elem
		i=-size
		u=i+structsz_obj_file_elem

		if not u:
			objinfo=rentfile[i:]
		else:
			objinfo=rentfile[i:u]

		result=unpack(structfmt_obj_file_elem, objinfo)

		# Do some surgery on this value..
		# It comes in as a `long`, but the format is 'I'
		# (for unsigned int); struct.pack() doesn't do
		# this reverse coercion.
		result=list(result)
		result[0]=int(result[0])

		yield tuple(result)

		size-=structsz_obj_file_elem

def get_rent(path):
	"Assemble rent-object from scan_rentfile()"
	scan=scan_rentfile(path)
	rent=[next(scan)]

	for obj in scan:
		rent.append(obj)

	return rent

def cstring(s):
	i = s.find('\x00')
	if i < 0:
		return s

	return s[:i]

class Item:
	"""
	Represents this binary packed structure:
	---
	   obj_vnum item_number;

	   char name[128];
	   char description[256];
	   char short_description[128];
	   sh_int locate;
	   int	value[4];
	   int  item_type;
	   int /*bitvector_t*/	extra_flags;
	   int  anti_flags;
	   int  wear_flags;
	   int	weight;
	   int	timer;
	   long /*bitvector_t*/	bitvector;
	   struct obj_affected_type affected[MAX_OBJ_AFFECT];
	"""

	def __init__(self, rent):
		self.vnum     = rent[0]

		self.keywords = cstring(rent[1])
		self.desc     = cstring(rent[2])
		self.name     = cstring(rent[3])

		self.savedAt  = rent[4]
		self.values   = rent[5:9]

		self.itemType = rent[9]

		self.extra    = rent[10]
		self.anti     = rent[11]
		self.wear     = rent[12]

		self.weight   = rent[13]
		self.timer    = rent[14]
		self.bitv     = rent[15]

		self.affect   = rent[16:]

		# self.rent = rent

	def __repr__(self):
		return '<Item #%d (%s)>' % (self.vnum, self.name)

	pf = staticmethod(__import__('pprint').pformat)

	def __str__(self):
		return self.pf(self.__dict__)
